#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This module holds all the user needs to deal with Volume of interests.
It provides the top-level VdxCube class, Voi, Slice and Contour classes.
The Voi class represents a volume of interest 'VOI', also called region of interest 'ROI' in Dicom lingo.
Each Voi holds several Slice, which are normally synced with an associated CT-cube.
Each Slice holds one or more Contours.
"""
import os
import copy
from math import pi, sqrt
import logging
import warnings
from functools import cmp_to_key

import numpy as np

try:
    # as of version 1.0 pydicom package import has beed renamed from dicom to pydicom
    from pydicom import uid
    from pydicom.dataset import Dataset, FileDataset
    from pydicom.sequence import Sequence
    _dicom_loaded = True
except ImportError:
    try:
        # fallback to old (<1.0) pydicom package version
        from dicom import UID as uid  # old pydicom had UID instead of uid
        from dicom.dataset import Dataset, FileDataset
        from dicom.sequence import Sequence
        _dicom_loaded = True
    except ImportError:
        _dicom_loaded = False

import pytrip
from pytrip.error import InputError, ModuleNotLoadedError
from pytrip.dos import DosCube
import pytriplib

logger = logging.getLogger(__name__)


class VdxCube:
    """
    VdxCube is the master class for dealing with Volume of Interests (VOIs).
    A VdxCube contains one or more VOIs which are structures which represent
    some organ (lung, eye ...) or target (GTV, PTV...)
    The Voi object contains Slice objects which corresponds to the CT slices,
    and the slice objects contains contour objects.
    Each contour object are a set of points which delimit a closed region.
    One single slice object can contain multiple contours.

    VdxCube ---> Voi[] ---> Slice[] ---> Contour[] ---> Point[]

    Note, since TRiP98 only supports one contour per slice for each voi.
    PyTRiP supports functions for connecting multiple contours to a single
    entity using infinite thin connects.

    VdxCube can import both dicom data and TRiP data,
    and export in the those formats.

    We strongly recommend to load a CT and/or a DOS cube first, see example below:

    >>> import pytrip as pt
    >>> c = pt.CtxCube()
    >>> c.read("tests/res/TST003/tst003000.ctx.gz")
    >>> v = pt.VdxCube(c)
    >>> v.read("tests/res/TST003/tst003000.vdx")
    """

    # stictly, VDX does not have .hed companion. However, in practice, .vdx files are always
    # associated with a .hed .ctx cube pair. In order to discover these, the header file
    # extentions may be associated to this as a .vdx data file.

    header_file_extension = '.hed'
    data_file_extension = '.vdx'
    allowed_suffix = tuple()

    def __init__(self, cube=None):
        """
        :param cube: CtxCube type object
        """
        self.vois = []
        self.cube = cube
        self.path = ""  # full path to .vdx file, set if a regular .vdx file was loaded loaded

        # UIDs unique for whole structure set
        # generation of UID is done here in init, the reason why we are not generating them in create_dicom
        # method is that subsequent calls to write method shouldn't changed UIDs
        self._dicom_study_instance_uid = uid.generate_uid(prefix=None)
        self._structs_dicom_series_instance_uid = uid.generate_uid(prefix=None)
        self._structs_sop_instance_uid = uid.generate_uid(prefix=None)
        self._structs_rt_series_instance_uid = uid.generate_uid(prefix=None)

        self.version = "2.0"
        if self.cube is not None:
            self.patient_id = cube.patient_id
            self._dicom_study_instance_uid = self.cube._dicom_study_instance_uid
            logger.debug("VDX class inherited patient_id {}".format(self.patient_id))
        else:
            import datetime
            self.patient_id = datetime.datetime.today().strftime('%Y%m%d-%H%M%S')
            logger.debug("VDX class creates new patient_id {}".format(self.patient_id))

    def __str__(self):
        """ str output handler
        """
        return self._print()

    def _print(self):
        """ Pretty print current attributes.
        """
        out = "\n"
        out += "   VdxCube\n"
        out += "----------------------------------------------------------------------------\n"
        out += "| UIDs\n"
        out += "|   dicom_study_instance_uid            : {:s}\n".format(self._dicom_study_instance_uid)
        out += "|   structs_dicom_series_instance_uid   : '{:s}'\n".format(self._structs_dicom_series_instance_uid)
        out += "|   structs_sop_instance_uid            : '{:s}'\n".format(self._structs_sop_instance_uid)
        out += "|   structs_rt_series_instance_uid      : '{:s}'\n".format(self._structs_rt_series_instance_uid)
        if self.vois:
            out += "+---VOIs\n"
            for _i, _v in enumerate(self.vois):
                out += "|   |           #{:d}              : '{:s}'\n".format(_i, _v.name)
        return out

    def read_dicom(self, data, structure_ids=None):
        """
        Reads structures from a Dicom RTSS Object.

        :param Dicom data: A Dicom RTSS object.
        :param structure_ids: (TODO: undocumented)
        """
        if "rtss" not in data:
            raise InputError("Input is not a valid rtss structure")
        dcm = data["rtss"]
        self.version = "2.0"

        self._dicom_study_instance_uid = dcm.StudyInstanceUID
        self._structs_dicom_series_instance_uid = dcm.SeriesInstanceUID
        self._structs_sop_instance_uid = dcm.SOPInstanceUID

        self.basename = dcm.PatientID.replace(" ", "_")

        if hasattr(dcm, 'ROIContourSequence'):
            _contours = dcm.ROIContourSequence
        else:
            logger.error("No ROIContours or ROIContourSequence found in dicom RTSTRUCT")
            exit()

        for i, _roi_contour in enumerate(_contours):
            if structure_ids is None or _roi_contour.RefdROINumber in structure_ids:
                if hasattr(dcm, "RTROIObservationsSequence"):
                    _roi_observation = dcm.RTROIObservationsSequence[i]
                    # _roi_name = dcm.RTROIObservationsSequence[i].ROIObservationLabel  # OPTIONAL by DICOM
                    # kept for future use.

                else:
                    logger.error("No RTROIObservations or RTROIObservationsSequence found in dicom RTSTRUCT")
                    exit()

                if hasattr(dcm, "StructureSetROISequence"):
                    _roi_name = dcm.StructureSetROISequence[i].ROIName  # REQUIRED by DICOM. At least an empty string.
                else:
                    logger.error("No StructureSetROISequence found in dicom RTSTRUCT")
                    exit()

                v = Voi(_roi_name, self.cube)
                v.read_dicom(_roi_observation, _roi_contour, _roi_name)

                self.add_voi(v)

    def get_voi_names(self):
        """
        :returns: a list of available voi names.
        :warning: deprecated method, use voi_names() instead.
        """
        warnings.warn("Call to deprecated method get_voi_names(). Use voi_names() instead.",
                      category=DeprecationWarning,
                      stacklevel=2)
        return self.voi_names()

    def voi_names(self):
        """
        :returns: a list of available voi names.
        """
        names = [voi.name for voi in self.vois]
        return names

    def add_voi(self, voi):
        """ Appends a new voi to this class.

        :param Voi voi: the voi to be appended to this class.
        """
        self.vois.append(voi)

    def get_voi_by_name(self, name):
        """ Returns a Voi object by its name.

        :param str name: Name of voi to be returned.
        :returns: the Voi which has exactly this name, else raise an Error.
        """
        for voi in self.vois:
            if voi.name.lower() == name.lower():
                return voi
        raise InputError("Voi doesn't exist")

    def import_vdx(self, path):
        """ Reads a structure file in Voxelplan format.
        This method is identical to self.read() and self.read_vdx()

        :param str path: Full path including file extension.
        """
        self.read_vdx(path)

    def read(self, path):
        """ Reads a structure file in Voxelplan format.
        This method is identical to self.import_vdx() and self.read_vdx()

        :param str path: Full path including file extension.
        """
        self.read_vdx(path)

    def vdx_version(self, content):
        """ Test content for what VDX version this is.
        Since the VDX version is very undocumented, we are guessing here.

        :param content: the vdx file contents as an array of strings.
        :returns: vdx version string, e.g. "1.2" or "2.0"
        """
        for line in content:
            if line.strip().startswith("vdx_file_version"):
                return line.split()[1]
            else:
                return "1.2"

    def read_vdx(self, path):
        """ Reads a structure file in Voxelplan format.

        :param str path: Full path including file extension.
        """
        self.basename = os.path.basename(path).split(".")[0]
        self.path = path
        fp = open(path, "r")
        content = fp.read().split('\n')
        fp.close()

        self.version = self.vdx_version(content)

        i = 0
        n = len(content)
        header_full = False
        #        number_of_vois = 0
        while i < n:
            line = content[i].strip()
            if not header_full:
                if line.startswith("all_indices_zero_based"):
                    self.zero_based = True
#                TODO number_of_vois not used
#                elif "number_of_vois" in line:
#                    number_of_vois = int(line.split()[1])
            if line.startswith("voi"):
                v = Voi(line.split()[1], self.cube)
                if self.version == "1.2":
                    _token = line.split()
                    if len(_token) == 6:
                        if _token[5] != '0':
                            i = v.read_vdx_old(content, i)
                else:
                    i = v.read_vdx(content, i)
                self.add_voi(v)
                header_full = True
            i += 1

    def concat_contour(self):
        """ Loop through all available VOIs and check whether any have multiple contours in a slice.
        If so, merge them to a single contour.

        This is needed since TRiP98 cannot handle multiple contours in the same slice.
        """
        for i in range(len(self.vois)):
            self.vois[i].concat_contour()

    def number_of_vois(self):
        """
        :returns: the number of VOIs in this object.
        """
        return len(self.vois)

    def _write_vdx(self, path):
        """ Writes all VOIs in voxelplan format.

        All will be written in version 2.0 format, irrespectively of self.version.

        :param str path: Full path, including file extension (.vdx).
        """
        fp = open(path, "w")
        fp.write("vdx_file_version 2.0\n")
        fp.write("all_indices_zero_based\n")
        fp.write("number_of_vois {:d}\n".format(self.number_of_vois()))
        self.vois = sorted(self.vois, key=lambda voi: voi.type, reverse=True)
        for voi in self.vois:
            logger.debug("writing VOI {}".format(voi.name))
            fp.write(voi.vdx_string())
        fp.close()

    def write_trip(self, path):
        """ Writes all VOIs in voxelplan format, while ensuring no slice holds more than one contour.
        Identical to write().

        :param str path: Full path, including file extension (.vdx).
        """
        self.concat_contour()
        self._write_vdx(path)

    def write(self, path):
        """ Writes all VOIs in voxelplan format, while ensuring no slice holds more than one contour.
        Identical to write_trip().

        :param str path: Full path, including file extension (.vdx).
        """
        self.write_trip(path)

    def create_dicom(self):
        """ Generates and returns Dicom RTSTRUCT object, which holds all VOIs.

        :returns: a Dicom RTSTRUCT object holding any VOIs.
        """
        if _dicom_loaded is False:
            raise ModuleNotLoadedError("Dicom")
        meta = Dataset()
        meta.MediaStorageSOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3'  # RT Structure Set Storage SOP Class
        # SOP Instance UID tag 0x0002,0x0003 (type UI - Unique Identifier)
        meta.MediaStorageSOPInstanceUID = self._structs_sop_instance_uid
        meta.ImplementationClassUID = "1.2.3.4"
        meta.TransferSyntaxUID = uid.ImplicitVRLittleEndian  # Implicit VR Little Endian - Default Transfer Syntax
        ds = FileDataset("file", {}, file_meta=meta, preamble=b"\0" * 128)
        if self.cube is not None:
            ds.PatientName = self.cube.patient_name
            ds.Manufacturer = self.cube.creation_info  # Manufacturer tag, 0x0008,0x0070 (type LO - Long String)
        else:
            ds.PatientName = ''
            ds.Manufacturer = ''  # Manufacturer tag, 0x0008,0x0070 (type LO - Long String)
        ds.SeriesNumber = '1'  # SeriesNumber tag 0x0020,0x0011 (type IS - Integer String)
        ds.PatientID = self.patient_id  # patient_id of the VdxCube, from CtxCube during __init__.
        ds.PatientSex = ''  # Patient's Sex tag 0x0010,0x0040 (type CS - Code String)
        #                      Enumerated Values: M = male F = female O = other.
        ds.PatientBirthDate = '19010101'
        ds.SpecificCharacterSet = 'ISO_IR 100'
        ds.AccessionNumber = ''
        ds.is_little_endian = True
        ds.is_implicit_VR = True
        ds.SOPClassUID = '1.2.840.10008.5.1.4.1.1.481.3'  # RT Structure Set Storage SOP Class

        # SOP Instance UID tag 0x0008,0x0018 (type UI - Unique Identifier)
        ds.SOPInstanceUID = self._structs_sop_instance_uid

        # Study Instance UID tag 0x0020,0x000D (type UI - Unique Identifier)
        # self._dicom_study_instance_uid may be either set in __init__ when creating new object
        #   or set when import a DICOM file
        #   Study Instance UID for structures is the same as Study Instance UID for CTs
        ds.StudyInstanceUID = self._dicom_study_instance_uid

        # Series Instance UID tag 0x0020,0x000E (type UI - Unique Identifier)
        # self._rt_dicom_series_instance_uid may be either set in __init__ when creating new object
        #   or set when import a DICOM file
        #   Series Instance UID for structures might be different than Series Instance UID for CTs
        ds.SeriesInstanceUID = self._structs_dicom_series_instance_uid

        ds.FrameOfReferenceUID = '1.2.3'  # !!!!!!!!!
        ds.SeriesDate = '19010101'  # !!!!!!!!
        ds.ContentDate = '19010101'  # !!!!!!
        ds.StudyDate = '19010101'  # !!!!!!!
        ds.StudyID = '1'  # Study ID tag 0x0020,0x0010 (type SH - Short String)
        ds.SeriesTime = '000000'  # !!!!!!!!!
        ds.StudyTime = '000000'  # !!!!!!!!!!
        ds.ContentTime = '000000'  # !!!!!!!!!
        ds.StructureSetLabel = 'PyTRiP structs'  # Structure set label tag, 0x3006,0x0002 (type SH - Short String)
        # Short string (SH) is limited to 16 characters !
        ds.StructureSetDate = '19010101'
        ds.StructureSetTime = '000000'
        ds.StructureSetName = 'ROI'
        ds.Modality = 'RTSTRUCT'
        ds.ROIGenerationAlgorithm = '0'  # ROI Generation Algorithm tag, 0x3006,0x0036 (type CS - Code String)
        ds.ReferringPhysicianName = 'py^trip'  # Referring Physician's Name tag 0x0008,0x0090 (type PN - Person Name)

        roi_label_list = []
        roi_data_list = []
        roi_structure_roi_list = []

        # to get DICOM which can be loaded in Eclipse we need to store information about UIDs of all slices in CT
        # first we check if DICOM cube is loaded
        if self.cube is not None:
            rt_ref_series_data = Dataset()
            rt_ref_series_data.SeriesInstanceUID = self.cube._ct_dicom_series_instance_uid
            rt_ref_series_data.ContourImageSequence = Sequence([])

            # each CT slice corresponds to one DICOM file
            for slice_dicom in self.cube.create_dicom():
                slice_dataset = Dataset()
                slice_dataset.ReferencedSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage SOP Class
                slice_dataset.ReferencedSOPInstanceUID = slice_dicom.SOPInstanceUID  # most important - slice UID
                rt_ref_series_data.ContourImageSequence.append(slice_dataset)

            rt_ref_study_seq_data = Dataset()
            rt_ref_study_seq_data.ReferencedSOPClassUID = '1.2.840.10008.3.1.2.3.2'  # Study Component Management Class
            rt_ref_study_seq_data.ReferencedSOPInstanceUID = '1.2.3'
            rt_ref_study_seq_data.RTReferencedSeriesSequence = Sequence([rt_ref_series_data])

            rt_ref_frame_study_data = Dataset()
            rt_ref_frame_study_data.RTReferencedStudySequence = Sequence([rt_ref_study_seq_data])
            rt_ref_frame_study_data.FrameOfReferenceUID = '1.2.3'
            # (3006, 0010) 'Referenced Frame of Reference Sequence'
            ds.ReferencedFrameOfReferenceSequence = Sequence([rt_ref_frame_study_data])

        for i in range(self.number_of_vois()):
            logger.debug("Write ROI #{:d} to DICOM object".format(i))
            roi_label = self.vois[i].create_dicom_label()
            roi_label.ObservationNumber = str(i + 1)
            roi_label.ReferencedROINumber = str(i + 1)
            roi_label.RefdROINumber = str(i + 1)
            roi_contours = self.vois[i].create_dicom_contour_data(i)
            roi_contours.RefdROINumber = str(i + 1)
            roi_contours.ReferencedROINumber = str(i + 1)

            roi_structure_roi = self.vois[i].create_dicom_structure_roi()
            roi_structure_roi.ROINumber = str(i + 1)

            # (3006, 0024) Referenced Frame of Reference UID   (UI)
            roi_structure_roi.ReferencedFrameOfReferenceUID = rt_ref_frame_study_data.FrameOfReferenceUID
            roi_structure_roi_list.append(roi_structure_roi)
            roi_label_list.append(roi_label)
            roi_data_list.append(roi_contours)
        ds.RTROIObservations = Sequence(roi_label_list)
        ds.ROIContours = Sequence(roi_data_list)
        ds.StructureSetROIs = Sequence(roi_structure_roi_list)
        return ds

    def write_dicom(self, directory):
        """ Generates a Dicom RTSTRUCT object from self, and writes it to disk.

        :param str directory: Directory where the rtss.dcm file will be saved.
        """
        dcm = self.create_dicom()
        dcm.save_as(os.path.join(directory, "RTSTRUCT.PYTRIP.dcm"))


def _voi_point_cmp(a, b):
    """ TODO: needs documentation """
    if abs(a[1] - b[1]) < 0.2:
        c = a[0] - b[0]
    else:
        c = a[1] - b[1]
    if c < 0:
        return -1
    else:
        return 1


def create_cube(cube, name, center, width, height, depth):
    """
    Creates a new VOI which holds the contours rendering a square box

    :param Cube cube: A CTX or DOS cube to work on.
    :param str name: Name of the VOI
    :param [float*3] center: Center position [x,y,z] in [mm]
    :param float width: Width of box, along x in [mm]
    :param float height: Height of box, along y in [mm]
    :param float depth: Depth of box, along z in [mm]
    :returns: A new Voi object.
    """
    v = Voi(name, cube)
    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - depth / 2 <= z <= center[2] + depth / 2:
            s = Slice(cube)
            s.thickness = cube.slice_distance
            points = [  # 4 corners of cube in this slice
                [center[0] - width / 2, center[1] - height / 2, z],
                [center[0] + width / 2, center[1] - height / 2, z],
                [center[0] + width / 2, center[1] + height / 2, z],
                [center[0] - width / 2, center[1] + height / 2, z]
            ]
            c = Contour(points, cube)
            c.contour_closed = True
            s.add_contour(c)
            v.add_slice(s)
    return v


def create_voi_from_cube(cube, name, value=100):
    """
    Creates a new VOI which holds the contours following an isodose lines.

    :param Cube cube: A CTX or DOS cube to work on.
    :param str name: Name of the VOI
    :param int value: The isodose value from which the contour will be generated from.
    :returns: A new Voi object.
    """
    v = Voi(name, cube)
    # there are some cases when this script is run on systems without DISPLAY variable being set
    # in such case matplotlib backend has to be explicitly specified
    # we do it here and not in the top of the file, as interleaving imports with code lines is discouraged
    import _cntr
    for i in range(cube.dimz):
        x, y = np.meshgrid(np.arange(len(cube.cube[0, 0])), np.arange(len(cube.cube[0])))
        isodose_obj = _cntr.Cntr(x, y, cube.cube[i])
        contour = isodose_obj.trace(value)
        s = Slice(cube)
        s.thickness = cube.slice_distance
        if not len(contour):
            continue
        points = np.zeros((len(contour[0]), 3))
        points[:, 0:2] = contour[0] * cube.pixel_size

        points[:, 2] = i * cube.slice_distance
        c = Contour(points.tolist(), cube)
        c.contour_closed = True  # TODO: Probably the last point is double here
        s.add_contour(c)

        v.add_slice(s)
    return v


def create_cylinder(cube, name, center, radius, depth):
    """
    Creates a new VOI which holds the contours rendering a cylinder along z

    :param Cube cube: A CTX or DOS cube to work on.
    :param str name: Name of the VOI
    :param [float*3] center: Center position of cylinder [x,y,z] in [mm]
    :param float radius: Radius of cylinder in [mm]
    :param float depth: Depth of cylinder, along z in [mm]
    :returns: A new Voi object.
    """
    v = Voi(name, cube)
    t = np.linspace(start=0, stop=2.0 * pi, num=99, endpoint=False)
    p = list(zip(center[0] + radius * np.cos(t), center[1] + radius * np.sin(t)))

    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - depth / 2 <= z <= center[2] + depth / 2:
            s = Slice(cube)
            s.thickness = cube.slice_distance
            points = [[x[0], x[1], z] for x in p]
            if points:
                c = Contour(points, cube)
                c.contour_closed = True
                s.add_contour(c)
                v.add_slice(s)
    return v


def create_sphere(cube, name, center, radius):
    """
    Creates a new VOI which holds the contours rendering a sphere along z

    :param Cube cube: A CTX or DOS cube to work on.
    :param str name: Name of the VOI
    :param [float*3] center: Center position of sphere [x,y,z] in [mm]
    :param float radius: Radius of sphere in [mm]
    :returns: A new Voi object.
    """
    v = Voi(name, cube)

    t = np.linspace(start=0, stop=2.0 * pi,
                    num=99, endpoint=False)  # num: sets the number of corners in sphere per slice.
    p = list(zip(np.cos(t), np.sin(t)))

    points = []

    for i in range(0, cube.dimz):
        z = i * cube.slice_distance
        if center[2] - radius <= z <= center[2] + radius:
            r2 = radius**2 - (z - center[2])**2
            s = Slice(cube)
            s.thickness = cube.slice_distance
            _contour_closed = True
            if r2 > 0.0:
                points = [[center[0] + x[0] * sqrt(r2),
                           center[1] + x[1] * sqrt(r2), z] for x in p]
            # in case r2 == 0.0, the contour in this slice is a point.
            # TODO: How should the sphere be treated with points in the end slices:
            # seen from the side: " .oOo. "  or should it be "  oOo  "  ?
            # The former means the voi consists of contours and points, which I am not sure is valid.
            # Here "  oOo  " is implemented.
            # If you do not want the " .oOo. " version uncomment the next three lines.
            else:
                points = [[center[0], center[1], z]]
            if len(points) > 0:
                c = Contour(points, cube)
                c.contour_closed = _contour_closed
                s.add_contour(c)
                v.add_slice(s)
    return v


class Voi:
    """
    This is a class for handling volume of interests (VOIs). This class can e.g. be found inside the VdxCube object.
    VOIs may for instance be organs (lung, eye...) or targets (PTV, GTV...), or any other volume of interest.
    """

    sagital = 2  #: deprecated, backwards compatibility to pytripgui, do not use.
    sagittal = 2  #: id for sagittal view
    coronal = 1  #: id for coronal view

    def __init__(self, name, cube=None):
        self.cube = cube
        self.name = name
        self.is_concated = False
        self.type = 90
        self.slices = []
        self.color = [0, 230, 0]  # default colour
        self.define_colors()

    def __str__(self):
        """ str output handler
        """
        return self._print()

    def _print(self):
        """ Pretty print current attributes.
        """
        out = "\n"
        out += "   Voi\n"
        out += "----------------------------------------------------------------------------\n"
        out += "|   Name                                : '{:s}'\n".format(self.name)
        out += "|   Is concated                         : {:s}\n".format(str(self.is_concated))
        out += "|   Type                                : {:d}\n".format(self.type)
        out += "|   Number of slices in VOI             : {:d}\n".format(len(self.slices))
        out += "|   Color 0xRGB                         : #{:s}{:s}{:s}\n".format(hex(self.color[0].strip('0x')),
                                                                                  hex(self.color[1].strip('0x')),
                                                                                  hex(self.color[2].strip('0x')))

        return out

    def create_copy(self, margin=0):
        """
        Returns an independent copy of the Voi object

        :param margin: (unused)
        :returns: a deep copy of the Voi object
        """
        voi = copy.deepcopy(self)
        if not margin == 0:
            pass
        return voi

    def get_voi_cube(self, level=1000, recalc=False):
        """
        This method returns a DosCube object with value 1000 in each voxel within the Voi and zeros elsewhere.
        It can be used as a mask, for selecting certain voxels.
        The function may take some time to execute the first invocation, but is faster for subsequent calls,
        due to caching.

        :param level: Level which will be set to every voxel which is inside VOI. 1000 by default.
        :param recalc: force recalculation (avoid caching)
        :returns: a DosCube object which holds the value <level> in those voxels which are inside the Voi.
        """
        # caching: checks if class has voi_cube attribute
        # this is needed for speedup.

        if not recalc and hasattr(self, "voi_cube"):
            _max = self.voi_cube.cube.max()
            _max_inv = 1 / _max
            if _max == level:
                return self.voi_cube
            else:
                return self.voi_cube * _max_inv * level

        self.voi_cube = DosCube(self.cube)
        self.voi_cube.mask_by_voi_all(self, level)
        return self.voi_cube

    def add_slice(self, slice):
        """ Add another slice to this VOI, and update self.slice_z table.

        :param Slice slice: the Slice object to be appended.
        """
        self.slices.append(slice)

    def get_name(self):
        """
        :returns: The name of this VOI.
        """
        return self.name

    def calculate_bad_angles(self, voi):
        """
        (Not implemented.)
        """
        pass

    def concat_to_3d_polygon(self):
        """ Concats all contours into a single contour, and writes all data points to self.polygon3d.
        """
        self.concat_contour()
        data = []
        for sl in self.slices:
            data.extend(sl.contours[0].contour)
        self.polygon3d = np.array(data)

    def get_3d_polygon(self):
        """ Returns a list of points rendering a 3D polygon of this VOI, which is stored in
        sefl.polygon3d. If this attribute does not exist, create it.
        """
        if not hasattr(self, "polygon3d"):
            self.concat_to_3d_polygon()
        return self.polygon3d

    def create_point_tree(self):
        """
        Concats all contours.
        Writes a list of points into self.points describing this VOI.
        """
        points = {}
        self.concat_contour()

        for sl in self.slices:  # TODO: should be sorted
            contour = sl.contours[0].contour
            p = {}
            for x in contour:
                p[x[0], x[1], x[2]] = []
            points.update(p)
        n_slice = len(self.slices)
        last_contour = None

        for i, sl in enumerate(self.slices):
            contour = sl.contours[0].contour
            n_points = len(contour)
            if i < n_slice - 1:
                next_contour = self.slices[i + 1].contours[0].contour
            else:
                next_contour = None
            for j, point in enumerate(contour):
                j2 = (j + 1) % (n_points - 2)
                point2 = contour[j2]
                points[(point[0], point[1], point[2])].append(point2)
                points[(point2[0], point2[1], point2[2])].append(point)
                if next_contour is not None:
                    point3 = pytrip.res.point.get_nearest_point(point, next_contour)
                    points[(point[0], point[1], point[2])].append(point3)
                    points[(point3[0], point3[1], point3[2])].append(point)
                if last_contour is not None:
                    point4 = pytrip.res.point.get_nearest_point(point, last_contour)
                    if point4 not in points[(point[0], point[1], point[2])]:
                        points[(point[0], point[1], point[2])].append(point4)
                        points[(point4[0], point4[1], point4[2])].append(point)
            last_contour = contour
        self.points = points

    def get_2d_projection_on_basis(self, basis, offset=None):
        """
        (TODO: Documentation)
        """
        a = np.array(basis[0])
        b = np.array(basis[1])
        self.concat_contour()
        bas = np.array([a, b])
        data = self.get_3d_polygon()
        product = np.dot(data, np.transpose(bas))

        compare = self.cube.pixel_size
        filtered = pytriplib.filter_points(product, compare / 2.0)
        filtered = np.array(sorted(filtered, key=cmp_to_key(_voi_point_cmp)))
        filtered = pytriplib.points_to_contour(filtered)
        product = filtered

        if offset is not None:
            offset_proj = np.array([np.dot(offset, a), np.dot(offset, b)])
            product = product[:] - offset_proj
        return product

    def get_2d_slice(self, plane, depth):
        """
        Gets a 2d Slice object from the contour in either sagittal or coronal plane.
        Contours will be concatenated.

        :param int plane: either self.sagittal or self.coronal
        :param float depth: position of plane
        :returns: a Slice object.
        """
        self.concat_contour()
        points1 = []
        points2 = []
        for _slice in self.slices:  # TODO: slices must be sorted first, but wouldnt they always be ?
            if plane is self.sagittal:
                point = sorted(
                    pytriplib.slice_on_plane(np.array(_slice.contours[0].contour), plane, depth), key=lambda x: x[1])
            elif plane is self.coronal:
                point = sorted(
                    pytriplib.slice_on_plane(np.array(_slice.contours[0].contour), plane, depth), key=lambda x: x[0])
            if len(point) > 0:
                points2.append(point[-1])
                if len(point) > 1:
                    points1.append(point[0])
        s = None
        if len(points1) > 0:
            points1.extend(reversed(points2))
            points1.append(points1[0])
            s = Slice(cube=self.cube)
            s.add_contour(Contour(points1, cube=self.cube))

        return s

    def define_colors(self):
        """
        Creates a list of default colours [R,G,B] in self.colours.
        """
        self.colors = []
        self.colors.append([0, 0, 255])
        self.colors.append([0, 128, 0])
        self.colors.append([0, 255, 0])
        self.colors.append([255, 0, 0])
        self.colors.append([0, 128, 128])
        self.colors.append([255, 255, 0])

    def calculate_center(self):
        """
        Calculates the center of gravity for the VOI.

        :returns: A numpy array[x,y,z] with positions in [mm]
        """
        if hasattr(self, "center_pos"):
            return self.center_pos
        self.concat_contour()
        tot_volume = 0.0
        center_pos = np.array([0.0, 0.0, 0.0])
        for _slice in self.slices:
            center, area = _slice.calculate_center()
            tot_volume += area
            center_pos += area * center
        if tot_volume > 0:
            self.center_pos = center_pos / tot_volume
            return center_pos / tot_volume
        else:
            self.center_pos = center_pos
            return center_pos

    def get_color(self, i=None):
        """
        :param int i: selects a colour, default if None.
        :returns: a [R,G,B] list.
        """
        if i is None:
            return self.color
        return self.colors[i % len(self.colors)]

    def set_color(self, color):
        """
        :param [3*int]: set a color [R,G,B].
        """
        self.color = color

    def create_dicom_label(self):
        """ Based on self.name and self.type, a Dicom ROI_LABEL is generated.

        :returns: a Dicom ROI_LABEL
        """
        roi_label = Dataset()
        roi_label.ROIObservationLabel = self.name
        roi_label.RTROIInterpretedType = self.get_roi_type_name(self.type)
        return roi_label

    def create_dicom_structure_roi(self):
        """ Based on self.name, an empty Dicom ROI is generated.

        :returns: a Dicom ROI.
        """
        roi = Dataset()
        roi.ROIName = self.name
        return roi

    def create_dicom_contour_data(self, i):
        """
        Based on self.slices, DICOM contours are generated for the DICOM ROI.

        :returns: Dicom ROI_CONTOURS
        """
        roi_contours = Dataset()
        contours = []

        dcmcube = None
        if self.cube is not None:
            dcmcube = self.cube.create_dicom()

        for _slice in self.slices:
            logger.info("Get contours from slice at {:10.3f} mm".format(_slice.get_position()))
            contours.extend(_slice.create_dicom_contours(dcmcube))

        roi_contours.Contours = Sequence(contours)
        roi_contours.ROIDisplayColor = self.get_color(i)

        return roi_contours

    def _sort_slices(self):
        """
        Sorts all slices stored in self for increasing z.
        This is needed for displaying Saggital and Coronal view.
        """
        # slice_in_frame is only given by VDX, and these are also the only frames which need to be sorted
        # it seems. DICOM apparently have proper structure already. Nonetheless, this function is also
        # applied to DICOM contours.

        if len(self.slices) > 0 and hasattr(self.slices[0], "slice_in_frame"):
            self.slices.sort(key=lambda _slice: _slice.slice_in_frame, reverse=True)

    def read_vdx_old(self, content, i):
        """
        Reads a single VOI from Voxelplan .vdx data from 'content', assuming a legacy .vdx format.
        VDX format 1.2.

        :params [str] content: list of lines with the .vdx content
        :params int i: line number to the list.
        :returns: current line number, after parsing the VOI.
        """

        logger.debug("Reading legacy 1.2 VDX format.")
        line = content[i]
        items = line.split()
        self.name = items[1]
        self.type = int(items[3])
        i += 1
        while i < len(content):
            line = content[i].strip()
            if line.startswith("voi"):
                break
            if line.startswith("slice#"):
                s = Slice(cube=self.cube)
                i = s.read_vdx_old(content, i)  # Slices in .vdx files start at 0
                if self.cube is not None:
                    for cont1 in s.contours:
                        for cont2 in cont1.contour:
                            _slice_number = int(cont2[2])
                            # bound checking
                            if _slice_number > self.cube.dimz:
                                logger.error("VDX slice number# {:d} exceeds dimension of CTX "
                                             "cube zmax={:d}".format(_slice_number, self.cube.dimz))
                                raise Exception("VDX file not compatible with CTX cube")

                            # cont2[2] holds slice number (starting in 1), translate it to absolute position in [mm]
                            cont2[2] = self.cube.slice_to_z(int(cont2[2]))
                if s.get_position() is None:
                    raise Exception("cannot calculate slice position")

                self.slices.append(s)

            if line.startswith("#TransversalObjects"):
                pass
                # slices = int(line.split()[1]) # TODO holds information about number of skipped slices
            i += 1

        self._sort_slices()
        return i - 1

    def read_vdx(self, content, i):
        """
        Reads a single VOI from Voxelplan .vdx data from 'content'.
        VDX format version 2.0

        :params [str] content: list of lines with the .vdx content
        :params int i: line number to the list.
        :returns: current line number, after parsing the VOI.
        """
        logger.debug("Parsing VDX format 2.0")

        line = content[i]
        self.name = ' '.join(line.split()[1:])
        number_of_slices = 10000
        i += 1
        while i < len(content):
            line = content[i].strip()
            if line.startswith("key"):
                self.key = line.split()[1]
            elif line.startswith("type"):
                self.type = int(line.split()[1])
            elif line.startswith("number_of_slices"):
                number_of_slices = int(line.split()[1])
            elif line.startswith("slice "):  # need that extra space to discriminate from "slice_in_frame"
                s = Slice(cube=self.cube)
                i = s.read_vdx(content, i)
                if s.get_position() is None:
                    raise Exception("cannot calculate slice position")
                if self.cube is None:
                    raise Exception("cube not loaded")
                self.slices.append(s)

            elif line.startswith("voi"):
                break
            elif len(self.slices) >= number_of_slices:
                break
            i += 1

        self._sort_slices()
        return i - 1

    def get_roi_type_number(self, type_name):
        """
        :returns: 1 if GTV or CTV, 10 for EXTERNAL, else 0.
        """
        if type_name == 'EXTERNAL':
            return 10
        elif type_name == 'AVOIDANCE':
            return 2
        elif type_name == 'ORGAN':
            return 0
        elif type_name == 'GTV':
            return 1
        elif type_name == 'CTV':
            return 1
        else:
            return 0

    def get_roi_type_name(self, type_id):
        """
        :returns: The type name of the ROI.
        """
        if type_id == 10:
            return "EXTERNAL"
        elif type_id == 2:
            return 'AVOIDANCE'
        elif type_id == 1:
            return 'CTV'
        elif type_id == 0:
            return 'ORGAN'
        return ''

    def read_dicom(self, roi_obs, roi_cont, roi_name="(none)"):
        """
        Reads a single ROI (= VOI) from a Dicom data set.

        :param roi_obs: dcm['rtss'].RTROIObservationsSequence[i]
        :param roi_cont: dcm['rtss'].ROIContourSequence[i]
        :param roi_name: the name of this ROI.
        """

        if not hasattr(roi_cont, "ContourSequence"):
            logger.warning("No DICOM (3006,0050) Contour Data found in "
                           "(3006,0039) ROIContourSequence[] for ROI:'{}'".format(roi_name))
            return

        _roi_type_name = roi_obs.RTROIInterpretedType
        self.type = self.get_roi_type_number(_roi_type_name)
        self.color = roi_cont.ROIDisplayColor

        contours = roi_cont.ContourSequence
        for i, contour in enumerate(contours):

            # get current slice position
            _z_pos = contour.ContourData[2]

            # if we have a new z_position, add a new slice object to self.slices
            if _z_pos not in [_slice.get_position() for _slice in self.slices]:
                logger.debug("VOI {}: Append new slice at z = {:f} cm to slices list:".format(self.name, _z_pos))
                sl = Slice(cube=self.cube)
                self.slices.append(sl)
            else:
                # lookup proper slice (just to be sure, should the contours come in random order)
                logger.debug("VOI {}: Found multi-contour at z = {} cm".format(self.name, _z_pos))
                sl = self.get_slice_at_pos(_z_pos)

            # append the contour data to the contour list of this slice
            sl.add_dicom_contour(contour)

            # 'CLOSED_PLANAR' indicates that the last point shall be connected to the first point,
            # where the first point is not repeated in the Contour Data.
            #
            # 'POINT' indicates that the contour is a single point, defining a specific location of significance.
            #
            # 'OPEN_PLANAR' indicates that the last vertex shall not be connected to the first point,
            # and that all points in Contour Data (3006,0050) shall be coplanar.
            #
            # 'OPEN_NONPLANAR' indicates that the last vertex shall not be connected to the first point,
            # and that the points in Contour Data (3006,0050) may be non-coplanar.
            # This can be used to represent objects best described by a single, possibly non-coplanar curve,
            # such as a brachytherapy applicator.
            #
            # Reference: https://www.dabsoft.ch/dicom/3/C.8.8.6.1/

            # Apparently DICOM may have single points as contours which can be marked as either POINT or CLOSED_PLANAR.
            # Here, we will let any contour which is less than 3 points be an open contour per definition.
            # example a CLOSED_PLANAR point:
            # SLICERRT: pinnacle3-9.9-phantom-imrt
            # - ROIContourSequence[2].ContourSequence[16].ContourGeometricType

            if contour.NumberOfContourPoints > 2 and contour.ContourGeometricType == 'CLOSED_PLANAR':
                sl.contours[-1].contour_closed = True
            else:
                sl.contours[-1].contour_closed = False

        self._sort_slices()

    def vdx_string(self):
        """
        Returns list of strings for this voi in voxelplan format, which can be written into a .vdx file.
        VDX format 2.0 only.

        :returns: a list of str holding the all lines needed for a Voxelplan formatted file.
        """
        if len(self.slices) == 0:
            return ""

        out = "\n"
        out += "voi %s\n" % (self.name.replace(" ", "_"))
        out += "key empty\n"
        out += "type %s\n" % self.type
        out += "\n"
        out += "contours\n"
        out += "reference_frame\n"
        out += " origin 0.000 0.000 0.000\n"
        out += " point_on_x_axis 1.000 0.000 0.000\n"
        out += " point_on_y_axis 0.000 1.000 0.000\n"
        out += " point_on_z_axis 0.000 0.000 1.000\n"
        out += "number_of_slices %d\n" % self.number_of_slices()
        out += "\n"
        i = 0

        for sl in self.slices:
            pos = sl.get_position()
            out += "slice {:d}\n".format(i)
            out += "slice_in_frame {:.3f}\n".format(pos)
            out += "thickness {:.3f} reference start_pos {:.3f} stop_pos {:.3f}\n".format(sl.thickness,
                                                                                          pos - 0.5 * sl.thickness,
                                                                                          pos + 0.5 * sl.thickness)
            out += "number_of_contours {:d}\n".format(sl.number_of_contours())
            out += sl.vdx_string()
            i += 1
        return out

    def get_row_intersections(self, pos):
        """
        For a given postion pos, returns a sorted list of x-coordinates where the y-coordinate intersects the contours.

        :param pos: a 3D position in the form [x,y,z] (mm)
        :returns: a sorted list of all x coordinates of all contours intersecting with the contours.

        TODO: could be made static
        TODO: could be made private
        """
        _slice = self.get_slice_at_pos(pos[2])
        if _slice is None:
            return None
        return np.sort(_slice.get_intersections(pos))

    def get_slice_at_pos(self, z):
        """
        Finds and returns a slice object found at position z [mm] (float).

        :param float z: slice position in absolute coordiantes (i.e. including any offsets)
        :returns: VOI slice at position z, z position may be approxiamte
        """

        _slice = [item for item in self.slices if np.isclose(item.get_position(), z,
                                                             atol=item.thickness * 0.5)]
        if len(_slice) == 0:
            logger.debug("could not find slice in get_slice_at_pos() at position {}".format(z))
            return None
        else:
            logger.debug("found slice at pos for z: {:.2f} mm, thickness {:.2f} mm".format(z, _slice[0].thickness))
            return _slice[0]

    def number_of_slices(self):
        """
        :returns: number of slices covered by this VOI.
        """
        return len(self.slices)

    def concat_contour(self):
        """ Concat all contours in all slices found in this VOI.
        """
        if not self.is_concated:
            for sl in self.slices:
                sl.concat_contour()
        self.is_concated = True

    def get_min_max(self):
        """ Set self.temp_min and self.temp_max if they dont exist.

        :returns: minimum and maximum x y coordinates in Voi.
        """
        temp_min, temp_max = None, None
        if hasattr(self, "temp_min"):
            return self.temp_min, self.temp_max
        for _slice in self.slices:
            if temp_min is None:
                temp_min, temp_max = _slice.get_min_max()
            else:
                min1, max1 = _slice.get_min_max()
                temp_min = pytrip.res.point.min_list(temp_min, min1)
                temp_max = pytrip.res.point.max_list(temp_max, max1)
        self.temp_min = temp_min
        self.temp_max = temp_max
        return temp_min, temp_max


class Slice:
    """
    The Slice class is specific for structures, and should not be confused with Slices extracted from CTX or DOS
    objects.
    """

    def __init__(self, cube=None):
        self.cube = cube
        self.contours = []  # list of contours in this slice

        # the slice positions are recorded, however the thickness
        # may be smaller than just the distance between two slices.
        # Therefore it is here set to some small non-zero value.
        self.thickness = 0.1  # assume some default value
        if cube is not None:
            self.thickness = cube.slice_distance

    def add_contour(self, contour):
        """ Adds a new 'contour' to the existing contours.

        :param Contour contour: the contour to be added.
        """
        self.contours.append(contour)

    def add_dicom_contour(self, dcm):
        """ Adds a Dicom CONTOUR to the existing list of contours in this Slice class.

        :param Dicom dcm: a Dicom CONTOUR object.
        """

        # do not apply any offset here, since everything is written in real world coordinates.
        _offset = [0.0, 0.0, 0.0]
        self.contours.append(
            Contour(pytrip.res.point.array_to_point_array(np.array(dcm.ContourData, dtype=float), _offset),
                    self.cube))
        # add the slice position to slice_in_frame which is needed later for sorting.
        self.slice_in_frame = self.contours[-1].contour[0][2]

    def get_position(self):
        """
        :returns: the position of this slice in [mm] including zoffset
        """
        if len(self.contours) == 0:
            return None
        return self.contours[0].contour[0][2]

    def get_intersections(self, pos):
        """
        For a given position <pos>, return a list of x-coordinates intersecting where the pos.y intersects all
        closed contours found in this slice.

        :params pos: a position in the form [x,y]
        :returns: a list of x-coordinates intersecting the y-coordinate found in pos[1].

        :warning: method name may change.
        TODO: rename this method. It could possibly be merged with get_x_intersections.
        TODO: could be made static
        TODO: could be made private
        """
        intersections = []
        for c in self.contours:
            # do not include open contours.
            if c.contour_closed:
                intersections.extend(pytrip.res.point.get_x_intersection(pos[1], c.contour))
        return intersections

    def calculate_center(self):
        """ Calculate the center position of all contours in this slice.

        :returns: a list of center positions [x,y,z] in [mm] for each contour found.
        """
        tot_area = 0.0
        center_pos = np.array([0.0, 0.0, 0.0])
        for contour in self.contours:
            center, area = contour.calculate_center()
            tot_area += area
            center_pos += area * center
        if tot_area > 0:
            return center_pos / tot_area, tot_area
        else:
            return center_pos, tot_area

    def read_vdx(self, content, i):
        """
        Reads a single Slice from Voxelplan .vdx data from 'content'.
        VDX format 2.0.

        :params [str] content: list of lines with the .vdx content
        :params int i: line number to the list.
        :returns: current line number, after parsing the VOI.
        """
        self.thickness = 0.1  # some small default value
        line = content[i]
        number_of_contours = 0
        i += 1
        while i < len(content):
            line = content[i].strip()
            if line.startswith("slice_in_frame"):
                self.slice_in_frame = float(line.split()[1])
            elif line.startswith("thickness"):
                items = line.split()
                self.thickness = float(items[1])
                logger.debug("Read VDX: thickness = {:f}".format(self.thickness))

                if len(items) == 7:
                    self.start_pos = float(items[4])
                    self.stop_pos = float(items[6])
                else:
                    self.start_pos = float(items[3])
                    self.stop_pos = float(items[5])

            elif line.startswith("number_of_contours"):
                number_of_contours = int(line.split()[1])
            elif line.startswith("contour"):
                c = Contour([], self.cube)
                i = c.read_vdx(content, i)
                self.add_contour(c)
                # TODO: not sure if multiple contours for the same ROI/VOI are allowed in VDX format.
            elif line.startswith("slice "):  # need that extra space to discriminate from "slice_in_frame"
                break
            elif len(self.contours) >= number_of_contours:
                break
            i += 1
        return i - 1

    def read_vdx_old(self, content, i):
        """
        Reads a single Slice from Voxelplan .vdx data from 'content'.
        VDX format 1.2.

        :params [str] content: list of lines with the .vdx content
        :params int i: line number to the list.
        :returns: current line number, after parsing the VOI.
        """

        # VDX cubes in version 1.2 do not hold any information on slice thicknesses.
        line1 = content[i].strip()
        line2 = content[i + 1].strip()
        line3 = content[i + 2].strip()

        if not line1.startswith("slice#"):
            return None
        if not line2.startswith("#points"):
            return None
        if not line3.startswith("points"):
            return None

        self.slice_in_frame = float(line1.split()[1])

        c = Contour([], self.cube)
        c.read_vdx_old(slice_number=self.slice_in_frame, xy_line=line3.split()[1:])
        self.add_contour(c)

        return i

    def create_dicom_contours(self, dcmcube):
        """
        Creates and returns a list of Dicom CONTOUR objects from self.

        :param dcmcube: TODO write me
        """

        # in order to get DICOM readable by Eclipse we need to connect each contour with CT slice
        # CT slices are identified by SOPInstanceUID
        # first we assume some default value if we cannot figure out CT slice info (i.e. CT cube is not loaded)
        ref_sop_instance_uid = '1.2.3'

        # then we check if CT cube is loaded
        if dcmcube is not None:
            candidates = [dcm for dcm in dcmcube if np.isclose(dcm.SliceLocation,
                                                               self.get_position())]
            if len(candidates) > 0:
                # finally we extract CT slice SOP Instance UID
                ref_sop_instance_uid = candidates[0].SOPInstanceUID

        contour_list = []
        for item in self.contours:
            con = Dataset()
            contour = []
            for p in item.contour:
                contour.extend([p[0], p[1], p[2]])
            con.ContourData = contour
            con.ContourGeometricType = 'CLOSED_PLANAR'
            con.NumberofContourPoints = item.number_of_points()
            cont_image_item = Dataset()
            cont_image_item.ReferencedSOPClassUID = '1.2.840.10008.5.1.4.1.1.2'  # CT Image Storage SOP Class
            cont_image_item.ReferencedSOPInstanceUID = ref_sop_instance_uid  # CT slice Instance UID
            con.ContourImageSequence = Sequence([cont_image_item])
            contour_list.append(con)
        return contour_list

    def vdx_string(self):
        """
        Returns list of strings for this SLICE in voxelplan format, which can be written into a .vdx file.
        VDX format 2.0 only.

        :returns: a list of str holding the slice information with the countour lines for a Voxelplan formatted file.
        """
        out = ""
        for i, cnt in enumerate(self.contours):
            out += "contour %d\n" % i
            out += "internal false\n"

            if cnt.number_of_points() == 1:  # Handle POIs
                out += "number_of_points {:d}\n".format(cnt.number_of_points())
            else:  # Handle ROIs
                out += "number_of_points {:d}\n".format(cnt.number_of_points() + 1)
            out += cnt.vdx_string()
            out += "\n"
        return out

    def number_of_contours(self):
        """
        :returns: number of contours found in this Slice object.
        """
        return len(self.contours)

    def concat_contour(self):
        """
        Concat all contours in this Slice object to a single contour.
        """
        for i in range(len(self.contours) - 1, 0, -1):
            self.contours[0].push(self.contours[i])
            self.contours.pop(i)
        self.contours[0].concat()

    def remove_inner_contours(self):
        """
        Removes any "holes" in the contours of this slice, thereby changing the topology of the contour.
        """
        for i in range(len(self.contours) - 1, 0, -1):
            self.contours[0].push(self.contours[i])
            self.contours.pop(i)
        self.contours[0].remove_inner_contours()

    def get_min_max(self):
        """
        Set self.temp_min and self.temp_max if they dont exist.

        :returns: minimum and maximum x y coordinates in Voi.
        """
        temp_min, temp_max = self.contours[0].get_min_max()
        for i in range(1, len(self.contours)):
            min1, max1 = self.contours[i].get_min_max()
            temp_min = pytrip.res.point.min_list(temp_min, min1)
            temp_max = pytrip.res.point.max_list(temp_max, max1)
        return temp_min, temp_max


class Contour:
    """
    Class for handling single Contours.
    The contour class holds a list of points self.contour = [[x0,y0,z0], [x1,y1,z1], ... [xn, yn, zn]] in millimeters.
    A contour can also be a single point (POI).
    A contour may be open or closed.
    """

    def __init__(self, contour, cube=None):
        self.cube = cube
        self.children = []
        self.contour = contour  # TODO: consider renaming this to 'data' or 'contour_data'
        # contour_closed: if set to True, the last point in the contour will be repeated when writing VDX files.
        self.contour_closed = False

    def push(self, contour):
        """
        Push a contour on the contour stack.

        :param Contour contour: a Contour object.
        """
        for i in range(len(self.children)):
            if self.children[i].contains_contour(contour):
                self.children[i].push(contour)
                return
        self.add_child(contour)

    def calculate_center(self):
        """
        Calculate the center for a single contour, and the area of a contour in 3 dimensions.

        :returns: Center of the contour [x,y,z] in [mm], area [mm**2] (TODO: to be confirmed)
        """
        points = np.array(self.contour + [self.contour[0]])  # connect the contour first and last point
        # its needed only for calculation of dxdy in Green's theorem below
        dx_dy = np.diff(points, axis=0)
        if abs(points[0, 2] - points[1, 2]) < 0.01:
            area = -np.dot(points[:-1, 1], dx_dy[:, 0])
            paths = (dx_dy[:, 0]**2 + dx_dy[:, 1]**2)**0.5
        elif abs(points[0, 1] - points[1, 1]) < 0.01:
            area = -np.dot(points[:-1, 2], dx_dy[:, 0])
            paths = (dx_dy[:, 0]**2 + dx_dy[:, 2]**2)**0.5
        elif abs(points[0, 0] - points[1, 0]) < 0.01:
            area = -np.dot(points[:-1, 2], dx_dy[:, 1])
            paths = (dx_dy[:, 1]**2 + dx_dy[:, 2]**2)**0.5
        total_path = np.sum(paths)

        if total_path > 0:
            center = np.array([np.dot(points[:-1, 0], paths) / total_path,
                               np.dot(points[:-1, 1], paths) / total_path,
                               points[0, 2]])
        else:
            center = np.array([np.sum(points[:-1, 0]),
                               np.sum(points[:-1, 1]),
                               points[0, 2]])

        return center, area

    def get_min_max(self):
        """
        :returns: The lowest x,y,z values and the highest x,y,z values found in this Contour object.
        """
        min_x = np.amin(np.array(self.contour)[:, 0])
        min_y = np.amin(np.array(self.contour)[:, 1])
        min_z = np.amin(np.array(self.contour)[:, 2])

        max_x = np.amax(np.array(self.contour)[:, 0])
        max_y = np.amax(np.array(self.contour)[:, 1])
        max_z = np.amax(np.array(self.contour)[:, 2])
        return [min_x, min_y, min_z], [max_x, max_y, max_z]

    def vdx_string(self):
        """
        Returns list of strings for this CONTOUR in voxelplan format, which can be written into a .vdx file.
        VDX format 2.0 only.

        :returns: an array of str holding the contour points needed for a Voxelplan formatted file.
        """

        # The vdx files require all contours to be mapped to a CT cube starting at (x,y) = (0,0)
        # However, z may be mapped directly as found in the dicom file by using z_tables in .hed
        out = ""
        for i, cnt in enumerate(self.contour):
            out += " %.4f %.4f %.4f %.4f %.4f %.4f\n" % (cnt[0] - self.cube.xoffset,
                                                         cnt[1] - self.cube.yoffset,
                                                         cnt[2],
                                                         0, 0, 0)

        # repeat the first point, to close the contour, if needed
        if self.contour_closed and len(self.contour) > 1:
            out += " %.4f %.4f %.4f %.4f %.4f %.4f\n" % (self.contour[0][0] - self.cube.xoffset,
                                                         self.contour[0][1] - self.cube.yoffset,
                                                         self.contour[0][2],
                                                         0, 0, 0)
        return out

    def read_vdx(self, content, i):
        """
        Reads a single Contour from Voxelplan .vdx data from 'content'.
        VDX format 2.0.

       Note:
        - in VDX files the last contour point is repeated if the contour is closed.
        - If we have a point of interest, the length is 1.
        - Length 2 and 3 should thus never occur in VDX files (assuming all contours are closed)

        :params [str] content: list of lines with the .vdx content
        :params int i: line number to the list.
        :returns: current line number, after parsing the VOI.
        """
        self.contour_closed = False
        set_point = False
        points = 0
        j = 0
        while i < len(content):
            line = content[i].strip()

            # skip any empty lines
            if line == "":
                i += 1  # go to next line
                continue

            if set_point:  # expect list of x,y,z points
                if j >= points:
                    break

                con_dat = line.split()
                if len(con_dat) < 3:  # point must be at least three dimensional
                    logger.warning(".vdx line {:d}: ignored, expected <x> <y> <z> positions".format(i))
                else:
                    self.contour.append([float(con_dat[0]) + self.cube.xoffset,
                                         float(con_dat[1]) + self.cube.yoffset,
                                         float(con_dat[2])])
                    j += 1  # increment point counter

            else:  # in case we do not have a point, some keyword may be found
                if line.startswith("internal_false"):
                    self.internal_false = True
                if line.startswith("number_of_points"):
                    points = int(line.split()[1])
                    set_point = True
            i += 1  # go to next line

        # check if the contour is closed
        # self.contour[:] holds the actual data points
        if len(self.contour) > 1:  # check if this is an actual contour, and not a POI
            # if first data point is the same as the last data point we have closed contour
            if self.contour[0] == self.contour[-1]:
                self.contour_closed = True
                # and trash the last data point
                del self.contour[-1]
            else:
                self.contour_closed = False

        return i - 1

    def read_vdx_old(self, slice_number, xy_line):
        """
        Reads a single Contour from Voxelplan .vdx data from 'content' and appends it to self.contour data
        VDX format 1.2.

        See also notes in read_vdx(), regarding the length of a contour.

        :params slice_number: list of numbers (as characters) with slice number
        :params xy_line: list of numbers (as characters) representing X and Y coordinates of a contour
        """
        self.contour_closed = False

        if self.cube is None:
            _pixel_size = 1.0
            _warn = "Reading contour data in .vdx in 1.2 format: no CTX cube associated, setting pixel_size to 1.0."
            logger.warning(_warn)
        else:
            _pixel_size = self.cube.pixel_size

        # and example of xy_line: 3021 4761 2994 4899 2916 5015
        xy_pairs = [xy_line[i:i + 2] for i in range(0, len(xy_line), 2)]  # make list of pairs
        for x, y in xy_pairs:
            # TRiP98 saves X,Y coordinates as integers, to get [mm] they needs to be divided by 16
            self.contour.append([_pixel_size * float(x) / 16.0,
                                 _pixel_size * float(y) / 16.0,
                                 float(slice_number)])

        # The legacy 1.2 VDX format does not discriminate between open or closed contours.
        # Therefore all contours read will be closed, except for POIs.
        # I checked VDX 1.2 files generated by TRiP and converted from dcm2trip, the last point
        # is not repeated here.
        if len(self.contour) > 1:  # check if this is an actual contour, and not a POI
            logging.debug("VDX 1.2 read, contour is closed")
            self.contour_closed = True
        else:
            logging.debug("VDX 1.2 read, contour is a POI")
            self.contour_closed = False

    def add_child(self, contour):
        """ (TODO: Document me)
        """
        remove_idx = []
        for i, child in enumerate(self.children):
            if contour.contains_contour(child):
                contour.push(child)
                remove_idx.append(i)
        remove_idx.sort(reverse=True)
        for i in remove_idx:
            self.children.pop(i)
        self.children.append(contour)

    def number_of_points(self):
        """
        :returns: Number of points in this Contour object.
        """
        return len(self.contour)

    def has_childs(self):
        """
        :returns: True or False, whether this Contour object has children.
        """
        if len(self.children) > 0:
            return True
        return False

    def print_child(self, level):
        """
        Print child to stdout.

        :param int level: (TODO: needs documentation)
        """
        for i, item in enumerate(self.children):
            print(level * '\t', )
            print(item.contour)
            self.item.print_child(level + 1)

    def contains_contour(self, contour):
        """
        :returns: True if contour in argument is contained inside self.
        """
        return pytrip.res.point.point_in_polygon(contour.contour[0][0], contour.contour[0][1], self.contour)

    def concat(self):
        """
        In case of multiple contours in the same slice, this method will concat them to a single contour.
        This is important for TRiP98 compatibility, as TRiP98 cannot handle multiple contours in the same slice of
        of the same VOI.
        """
        for i in range(len(self.children)):
            self.children[i].concat()
        while len(self.children) > 1:
            d = -1
            child = 0
            for i in range(1, len(self.children)):
                i1_temp, i2_temp, d_temp = pytrip.res.point.short_distance_polygon_idx(
                    self.children[0].contour, self.children[i].contour)
                if d == -1 or d_temp < d:
                    d = d_temp
                    child = i
            i1_temp, i2_temp, d_temp = pytrip.res.point.short_distance_polygon_idx(
                self.children[0].contour, self.contour)
            if d_temp < d:
                self._merge(self.children[0])
                self.children.pop(0)
            else:
                self.children[0]._merge(self.children[child])
                self.children.pop(child)
        if len(self.children) == 1:
            self._merge(self.children[0])
            self.children.pop(0)

    def remove_inner_contours(self):
        """ (TODO: needs documentation)
        """
        for i in range(len(self.children)):
            self.children[i].children = []

    def _merge(self, contour):
        """
        Merge two contours into a single one.
        """
        if len(self.contour) == 0:
            self.contour = contour.contour
            return
        i1, i2, d = pytrip.res.point.short_distance_polygon_idx(self.contour, contour.contour)
        con = []
        for i in range(i1 + 1):
            con.append(self.contour[i])
        for i in range(i2, len(contour.contour)):
            con.append(contour.contour[i])
        for i in range(i2 + 1):
            con.append(contour.contour[i])
        for i in range(i1, len(self.contour)):
            con.append(self.contour[i])
        self.contour = con
        return



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This file contains the Execute() class which can execute a Plan() object locally or on a remote server
with a complete TRiP98 installations.
"""
import os
import uuid
import shutil
import tarfile
import logging

from subprocess import Popen, PIPE

from pytrip.dos import DosCube
from pytrip.let import LETCube
from pytrip.ctx import CtxCube

logger = logging.getLogger(__name__)

try:
    import paramiko
except ImportError:
    logger.warning("Paramiko package not installed, only local TRiP access possible")


class Execute(object):
    """ Execute class for running trip using attached Ctx, Vdx and Plan objects.
    """
    def __init__(self, ctx, vdx, ctx_path="", vdx_path=""):
        """ Initialize the Execute class.
        :params CtxCube() ctx: the CT images as a regular pytrip.CtxCube() object.
        :params VdxCube() ctx: the contours as a regular pytrip.VdxCube() object.

        If the Ctx and Vdx obejcts are constant, this object can be used to execute multiple
        plans in parallel.

        pytrip will write the Ctx and Vdx cube to a temporary directory where all will
        be dealt with. However, as it is right now, PyTRiP98 has sometimes trouble
        to convert to something which is readable TRiP98.
        For this case, we provide these optional hooks, where paths to Ctx and Vdx files can be
        specified, these will then be copied directly without any internal conversion.
        :params str ctx_path: optional source of .ctx cube, will override ctx.write()
        :params str vdx_path: optional source of .vdx cube, will override vdx.write()
        """

        logger.debug("Initializing TripExecuter()")

        self.__uuid__ = uuid.uuid4()  # for uniquely identifying this execute object

        self.ctx = ctx
        self._ctx_path = ctx_path

        self.vdx = vdx
        self._vdx_path = vdx_path

        self.listeners = []

        # TODO: this should be only the TRiP98 command, and not the full path
        # however some issue when running remote, so lets leave this for now
        self.trip_bin_path = "TRiP98"  # where TRiP98 is installed, if not accessible in /usr/local/bin or similar

        self.logfile_prefix_stdout = "trip98.stdout."  # will be suffixed by plan.basename
        self.logfile_prefix_stderr = "trip98.stderr."  # will be suffixed by plan.basename

        self._norun = False  # set this to False for a dry run without TRiP98 (for testing purposes)
        self._cleanup = True  # delete any temporary dirs created (currently this is only the package dir)

        # remote planning
        # remote directory where a new temporary directory will be created, the package extracted and executed.
        self.remote = False  # remote or local execution
        self.servername = ""
        self.username = ""
        self.password = ""
        self.remote_base_dir = "./"
        self.rsakey_local_path = "~/.ssh/id_rsa"

    def __str__(self):
        """ str output handler
        """
        return self._print()

    def _print(self):
        """ Pretty print current attributes.
        """
        out = "\n"
        out += "   Executer\n"
        out += "----------------------------------------------------------------------------\n"
        out += "| General configuration\n"
        out += "|   UUID                        : {:s}\n".format(str(self.__uuid__))
        out += "|   CtxCube.basename            : '{:s}'\n".format(self.ctx.basename)
        out += "|   VdxCube.basename            : '{:s}'\n".format(self.vdx.basename)
        out += "|   STDOUT prefix               : '{:s}'\n".format(self.logfile_prefix_stdout)
        out += "|   STDERR prefix               : '{:s}'\n".format(self.logfile_prefix_stderr)
        out += "|   TRiP98 command              : '{:s}'\n".format(self.trip_bin_path)
        out += "|   Cleanup                     : {:s}\n".format(str(self._cleanup))

        out += "|\n"
        out += "| Remote access\n"
        out += "|   Remote execution            : {:s}\n".format(str(self.remote))
        out += "|   Server                      : '{:s}'\n".format(self.servername)
        out += "|   Username                    : '{:s}'\n".format(self.username)
        out += "|   Password                    : '{:s}'\n".format("*" * len(self.password))

        return out

    def execute(self, plan, run=True, _callback=None):
        """
        Executes the Plan() object using TRiP98.

        :returns int: return code from trip execution.
        """

        if run:
            logger.debug("Execute TRiP98...")
            self._norun = False
        else:
            logger.debug("Execute TRiP98 (dry-run)...")
            self._norun = True

        self._callback = _callback  # TODO: check if GUI really needs this.
        self._pre_execute(plan)  # prepare directory where all will be run, put files in it.
        plan.save_exec(plan._exec_path)  # add the .exec as well
        rc = self._run_trip(plan)  # run TRiP
        self._finish(plan)
        return rc

    def _pre_execute(self, plan):
        """
        Prepare a temporary working directory where TRiP will be executed.
        Sets:
        _temp_dir where all will be executed
        _exec_path generates full path to the file name.exec file will be stored
        """

        # attach working dir to plan, but expanded for environment variables
        if not plan.working_dir:
            logger.warning("No working directory was specified for plan. Setting it to ./")
            plan.working_dir = "./"
        plan._working_dir = os.path.expandvars(plan.working_dir)

        # We will not run TRiP98 in working dir, but in an isolated subdirectory which will be
        # uniquely created for this purpose, located after the working dir.
        # This contents of this subdir will be termed "the package", as it is a clean environment.
        # The name of the subdir will be stored in plan._temp_dir
        # It may be safely deleted afterwards, once the resulting files are copied back to the plan.working_dir

        if not hasattr(plan, "_temp_dir"):
            import tempfile
            plan._temp_dir = tempfile.mkdtemp(prefix='trip98_', dir=plan._working_dir)

        plan._temp_dir = os.path.join(plan._working_dir,
                                      plan._temp_dir)

        plan._exec_path = os.path.join(plan._temp_dir,
                                       plan.basename + ".exec")

        logger.debug("Created temporary working directory {:s}".format(plan._temp_dir))

        _flist = []  # list of files which must be copied to the package.

        if plan.incube_basename:
            _flist.append(os.path.join(plan._working_dir, plan.incube_basename + ".dos"))
            _flist.append(os.path.join(plan._working_dir, plan.incube_basename + ".hed"))

        for _field in plan.fields:
            if _field.use_raster_file:
                _flist.append(os.path.join(plan._working_dir, _field.basename + ".rst"))

        # once the file list _flist is complete, copy it to the package location
        for _fn in _flist:
            logger.debug("Copy {:s} to {:s}".format(_fn, plan._temp_dir))
            shutil.copy(_fn, plan._temp_dir)

        # Ctx and Vdx files are not copied, but written from the objects passed to Execute() during __init__.
        # This gives the user better control over what Ctx and Vdx should be based for the planning.
        # Copying can though be forced by specifying .ctx and .vdx path during self.__init__
        # This is necessary as long as PyTRiP has trouble to produce TRiP98 readable files reliably.
        if self._ctx_path:
            _ctx_base, _ = os.path.splitext(self._ctx_path)
            logger.info("Copying {:s} to tmp dir, instead of writing from Ctx object.".format(self._ctx_path))
            shutil.copy(self._ctx_path, plan._temp_dir)  # copy .ctx
            shutil.copy(_ctx_base + ".hed", plan._temp_dir)  # copy.hed
        else:
            self.ctx.write(os.path.join(plan._temp_dir, self.ctx.basename + ".ctx"))  # will also make the hed

        if self._vdx_path:
            logger.info("Copying {:s} to tmp dir, instead of writing from Vdx object.".format(self._vdx_path))
            shutil.copy(self._vdx_path, plan._temp_dir)
        else:
            self.vdx.write(os.path.join(plan._temp_dir, self.vdx.basename + ".vdx"))

    def add_log_listener(self, listener):
        """ A listener is something which has a .write(txt) method.
        """
        self.listeners.append(listener)

    def log(self, txt):
        """ Writes txt to all listeners, stripping any newlines.
        """
        txt = txt.replace("\n", "")
        for l in self.listeners:
            l.write(txt)

    def _run_trip(self, plan, _dir=""):
        """ Method for executing the attached exec.
        :params str dir: overrides dir where the TRiP package is assumed to be, otherwise
        the package location is taken from plan._temp_dir

        :returns int: return code from trip execution.
        """
        if not _dir:
            _dir = plan._temp_dir
        if self.remote:
            rc = self._run_trip_remote(plan, _dir)
        else:
            rc = self._run_trip_local(plan, _dir)
        return rc

    def _run_trip_local(self, plan, _run_dir):
        """
        Runs TRiP98 on local computer.
        :params Plan plan : plan object to be executed
        :params str _run_dir: directory where a clean trip package is located, i.e. where TRiP98 will be run.

        :returns int: return code of TRiP98 execution. (0, even if optimiztion fails)
        """
        logger.info("Run TRiP98 in LOCAL mode.")

        trip, ver = self.test_local_trip()
        if trip is None:
            logger.error("Could not find TRiP98 using path \"{:s}\"".format(self.trip_bin_path))
            raise EnvironmentError
        else:
            logger.info("Found {:s} version {:s}".format(trip, ver))

        # stdout and stderr are always written locally
        _stdout_path = os.path.join(plan._working_dir, self.logfile_prefix_stdout + plan.basename)
        _stderr_path = os.path.join(plan._working_dir, self.logfile_prefix_stderr + plan.basename)
        logger.debug("Write stdout to {:s}".format(_stdout_path))
        logger.debug("Write stderr to {:s}".format(_stderr_path))

        fp_stdout = open(_stdout_path, "w")
        fp_stderr = open(_stderr_path, "w")

        _pwd = os.getcwd()
        os.chdir(_run_dir)

        # start local process running TRiP98
        if self._norun:  # for testing, just echo the command which would be executed
            p = Popen(["echo", self.trip_bin_path], stdout=PIPE, stdin=PIPE)
        else:
            p = Popen([self.trip_bin_path], stdout=PIPE, stdin=PIPE)

        # fill standard input with configuration file content
        # wait until process is finished and get standard output and error streams
        stdout, stderr = p.communicate(plan._trip_exec.encode("ascii"))
        rc = p.returncode
        if rc != 0:
            logger.error("TRiP98 error: return code {:d}".format(rc))
        else:
            logger.debug("TRiP98 exited with status: {:d}".format(rc))

        if stdout is not None:
            logger.debug("Local answer stdout:" + stdout.decode("ascii"))
            fp_stdout.write(stdout.decode("ascii"))
            self.log(stdout.decode("ascii"))

        if stderr is not None:
            logger.debug("Local answer stderr:" + stderr.decode("ascii"))
            fp_stderr.write(stderr.decode("ascii"))

        os.chdir(_pwd)

        fp_stdout.close()
        fp_stderr.close()

        return rc

    def _run_trip_remote(self, plan, _run_dir=None):
        """ Method for executing the attached plan remotely.
        :params Plan plan: plan object
        :params str _run_dir: TODO: not used

        :returns: integer exist staus of TRiP98 execution. (0, even if optimiztion fails)
        """
        logger.info("Run TRiP98 in REMOTE mode.")

        tar_path = self._compress_files(plan._temp_dir)  # make a tarball out of the TRiP98 package

        # prepare relevant dirs and paths
        _, tgz_filename = os.path.split(tar_path)
        remote_tgz_path = os.path.join(self.remote_base_dir, tgz_filename)
        local_tgz_path = os.path.join(plan._working_dir, tgz_filename)
        remote_run_dir = os.path.join(self.remote_base_dir, tgz_filename.rstrip(".tar.gz"))  # TODO: os.path.splitext
        remote_exec_fn = plan.basename + ".exec"
        remote_rel_run_dir = tgz_filename.rstrip(".tar.gz")  # TODO: os.path.splitext

        # stdout and stderr are always written locally
        _stdout_path = os.path.join(plan._working_dir, self.logfile_prefix_stdout + plan.basename)
        _stderr_path = os.path.join(plan._working_dir, self.logfile_prefix_stderr + plan.basename)
        logger.debug("Write stdout to {:s}".format(_stdout_path))
        logger.debug("Write stderr to {:s}".format(_stderr_path))

        # cp tarball to server
        self._copy_file_to_server(tar_path, remote_tgz_path)

        if self._norun:
            norun = "echo "
        else:
            norun = ""

        # The TRiP98 command must be encapsulated in a bash -l -c "TRiP98 < fff.exec"
        # then .bashrc_profile is checked. (However, not .bashrc)
        _tripcmd = "bash -l -c \"" + self.trip_bin_path + " < " + remote_exec_fn + "\""

        commands = ["cd " + self.remote_base_dir + ";" + "tar -zxvf " + remote_tgz_path,  # unpack tarball
                    "cd " + remote_run_dir + ";" + norun + _tripcmd,
                    "cd " + self.remote_base_dir + ";" + "tar -zcvf " + remote_tgz_path + " " + remote_rel_run_dir,
                    "cd " + self.remote_base_dir + ";" + "rm -r " + remote_run_dir]

        # test if TRiP is installed
        logger.debug("Test if TRiP98 can be reached remotely...")
        trip, ver = self.test_remote_trip()
        if trip is None:
            logger.error("Could not find TRiP98 on {:s} using path \"{:s}\"".format(self.servername,
                                                                                    self.trip_bin_path))
            raise EnvironmentError
        else:
            logger.info("Found {:s} version {:s} on {:s}".format(trip, ver, self.servername))

        fp_stdout = open(_stdout_path, "w")
        fp_stderr = open(_stderr_path, "w")

        # open ssh channel and run commands
        ssh = self._get_ssh_client()
        for _cmd in commands:
            logger.debug("Execute on remote server: {:s}".format(_cmd))
            self.log(_cmd)
            stdin, stdout, stderr = ssh.exec_command(_cmd)
            rc = int(stdout.channel.recv_exit_status())  # recv_exit_status() returns an string type
            if rc != 0:
                logger.error("TRiP98 error: return code {:d}".format(rc))
            else:
                logger.debug("TRiP98 exited with status: {:d}".format(rc))

            answer_stdout = stdout.read().decode('utf-8')
            answer_stderr = stderr.read().decode('utf-8')
            logger.info("Remote answer stdout:\n{:s}".format(answer_stdout))
            logger.info("Remote answer stderr:\n{:s}".format(answer_stderr))
            fp_stdout.write(answer_stdout)
            fp_stderr.write(answer_stderr)
            self.log(answer_stdout)

        fp_stdout.close()
        fp_stderr.close()
        ssh.close()

        self._move_file_from_server(remote_tgz_path, local_tgz_path)
        self._extract_tarball(remote_tgz_path, plan._working_dir)
        logger.debug("Locally remove {:s}".format(local_tgz_path))
        os.remove(local_tgz_path)

        return rc

    def _finish(self, plan):
        """ return requested results, copy them back in to plan._working_dir
        """

        for _file_name in plan._out_files:
            _path = os.path.join(plan._temp_dir, _file_name)

            # only copy files back, if we actually have been running TRiP
            if self._norun:
                logger.info("dummy run: would now copy {:s} to {:s}".format(_path, plan._working_dir))
            else:
                logger.info("copy {:s} to {:s}".format(_path, plan._working_dir))
                shutil.copy(_path, plan._working_dir)

        for _file_name in plan._out_files:
            _path = os.path.join(plan._temp_dir, _file_name)
            if ".phys.dos" in _file_name:
                _ctx_cube = DosCube()
                if not self._norun:
                    _ctx_cube.read(_path)
                    plan.dosecubes.append(_ctx_cube)

            if ".bio.dos" in _file_name:
                _ctx_cube = CtxCube()
                if not self._norun:
                    _ctx_cube.read(_path)
                    plan.dosecubes.append(_ctx_cube)

            if ".dosemlet.dos" in _file_name:
                _let_cube = LETCube()
                if not self._norun:
                    _let_cube.read(_path)
                    plan.letcubes.append(_let_cube)

            if ".rst" in _file_name:
                logger.warning("attaching fields to class not implemented yet {:s}".format(_path))
                # TODO
                # need to access the RstClass here for each rst file. This will then need to be attached
                # to the proper field in the list of fields.

        if self._cleanup:
            logger.debug("Delete {:s}".format(plan._temp_dir))
            shutil.rmtree(plan._temp_dir)

    @staticmethod
    def _compress_files(_source_dir):
        """
        Builds the tar.gz from what is found in _source_dir.
        Resulting file will be stored in "source_dir/.." if not specified otherwise

        :params str _source_dir: path to dir with the files to be compressed.
        :returns: full path to tar.gz file.

        :example:
        _compress_files("/home/bassler/test/foobar")
        will compress the contents of /foobar into foobar.tar.gz, including the foobar root dir
        and returns "/home/bassler/test/foobar.tar.gz"
        """

        # _source_dir may be of either "/foo/bar" or "/foo/bar/". We only need "bar",
        # but dirname() will return foo in the first case and bar in the second,
        # fix this by adding an extra seperator, in that case all will be stripped.
        # add an extra trailing slash, if there are multiple
        _dir = os.path.dirname(_source_dir + os.sep)
        _pardir, _basedir = os.path.split(_dir)
        _target_filename = _basedir + ".tar.gz"

        _cwd = os.getcwd()
        os.chdir(_pardir)

        logger.debug("Compressing files in {:s} to {:s}".format(_source_dir, _target_filename))

        with tarfile.open(_target_filename, "w:gz") as tar:
            tar.add(_basedir, arcname=_basedir)

        os.chdir(_cwd)

        return os.path.join(_pardir, _target_filename)

    @staticmethod
    def _extract_tarball(tgz_path, basedir):
        """ Extracts a tarball at path into self.working_dir
        :params tgz_path: full path to tar.gz file
        :params basedir: where extracted files (including the tgz root dir) will be stored
        :returns: a string where the files were extracted including the tgz root dir

        i.e. _extract_tarball("foobar.tar.gz", "/home/bassler/test")
        will return "/home/bassler/test/foobar"

        """

        logger.debug("Locally extract {:s} in {:s}".format(tgz_path, basedir))

        _basedir, _tarfile = os.path.split(tgz_path)
        _outdirname = _tarfile.rstrip(".tar.gz")  # TODO: os.path.splitext()
        _outdir = os.path.join(_basedir, _outdirname)

        if os.path.exists(_outdir):
            shutil.rmtree(_outdir)

        _cwd = os.getcwd()
        os.chdir(basedir)
        with tarfile.open(tgz_path, "r:gz") as tar:
            tar.extractall("./")

        os.chdir(_cwd)
        return _outdir

    def _copy_file_to_server(self, _from, _to):
        """
        Copies the generated tar.gz file to the remote server.
        :params _from: full path to file
        :params _to: full path to location on server
        """

        _to_uri = self.servername + ":" + _to
        logger.debug("Copy {:s} to {:s}".format(_from, _to_uri))

        sftp, transport = self._get_sftp_client()
        sftp.put(localpath=_from, remotepath=_to)
        sftp.close()
        transport.close()

    def _copy_file_from_server(self, _from, _to):
        """ Copies a single file from a server
        :param _from: full path on remote server
        :param _to: full path on local computer
        """
        _from_uri = self.servername + ":" + _from
        logger.debug("Copy {:s} to {:s}".format(_from_uri, _to))

        sftp, transport = self._get_sftp_client()
        sftp.get(_from, _to)
        sftp.close()
        transport.close()

    def _move_file_from_server(self, _from, _to):
        """ Copies a removes a single file from a server
        :param _from: full path on remote server
        :param _to: full path on local computer
        """
        _from_uri = self.servername + ":" + _from
        logger.debug("Move {:s} to {:s}".format(_from_uri, _to))

        sftp, transport = self._get_sftp_client()
        sftp.get(_from, _to)
        sftp.remove(_from)
        sftp.close()
        transport.close()

    def _get_ssh_client(self):
        """ returns an open ssh client
        """

        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # If no password is supplied, try to look for a private key
        if not self.password:
            rsa_keypath = os.path.expanduser(self.rsakey_local_path)
            if not os.path.isfile(rsa_keypath):
                # login with provided username + empty password
                try:
                    ssh.connect(self.servername, username=self.username, password="")
                except Exception:
                    logger.error("Cannot connect to " + self.servername)
                    logger.error("Check username, password or key in " + self.rsakey_local_path)
                    raise
            else:
                # login with provided username + private key
                rsa_key = paramiko.RSAKey.from_private_key_file(rsa_keypath)
                try:
                    ssh.connect(self.servername, username=self.username, pkey=rsa_key)
                except Exception:
                    logger.error("Cannot connect to " + self.servername)
                    logger.error("Check username and your key in " + self.rsakey_local_path)
                    raise
        else:
            # login with provided username + password
            ssh.connect(self.servername, username=self.username, password=self.password)

        return ssh

    def _get_sftp_client(self):
        """ returns a sftp client object and the corresponding transport socket.
        Both must be closed after use.
        """
        transport = paramiko.Transport((self.servername, 22))
        # transport.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # If no password is supplied, try to look for a private key
        if not self.password:
            rsa_keypath = os.path.expanduser(self.rsakey_local_path)
            if not os.path.isfile(rsa_keypath):
                # login with provided username + empty password
                try:
                    transport.connect(username=self.username, password="")
                except Exception:
                    logger.error("Cannot connect to " + self.servername)
                    logger.error("Check username, password or key in " + self.rsakey_local_path)
                    raise
            else:
                # login with provided username + private key
                rsa_key = paramiko.RSAKey.from_private_key_file(rsa_keypath)
                try:
                    transport.connect(username=self.username, pkey=rsa_key)
                except Exception:
                    logger.error("Cannot connect to " + self.servername)
                    logger.error("Check username and your key in " + self.rsakey_local_path)
                    raise
        else:
            # login with provided username + password
            transport.connect(username=self.username, password=self.password)

        sftp = paramiko.SFTPClient.from_transport(transport)

        return sftp, transport

    def test_local_trip(self):
        """ Test if TRiP98 can be reached locally.
        :returns tripname, tripver: Name of TRiP98 installation and its version.
        :returns: None, None if not installed
        """

        p = Popen([self.trip_bin_path], stdout=PIPE, stdin=PIPE)
        stdout, stderr = p.communicate("exit".encode('ascii'))

        _out = stdout.decode('utf-8')

        if "This is TRiP98" in _out:
            tripname = _out.split(" ")[3][:-1]
            tripver = _out.split(" ")[5].split("(")[0]
            return tripname, tripver
        else:
            return None, None

    def test_remote_trip(self):
        """ Test if TRiP98 can be reached remotely.
        :returns tripname, tripver: Name of TRiP98 installation and its version.
        :returns: None, None if not installed
        """

        # execute the list of commands on the remote server
        _tripcmd_test = "bash -l -c \"" + "cat exit | " + self.trip_bin_path + "\""

        # test if TRiP98 can be reached

        ssh = self._get_ssh_client()
        stdin, stdout, stderr = ssh.exec_command(_tripcmd_test)
        _out = stdout.read().decode('utf-8')
        ssh.close()

        if "This is TRiP98" in _out:
            tripname = _out.split(" ")[3][:-1]
            tripver = _out.split(" ")[5].split("(")[0]
            return tripname, tripver
        else:
            return None, None



#!/usr/bin/env python
#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Script for inspecting SPC files and producing PDF with summary plots
"""
import argparse
import datetime
import getpass
import logging
import math
import sys

import numpy as np

import pytrip as pt
from pytrip import spc


def main(args=sys.argv[1:]):

    # matplotlib on python3.2 (which is already deprecated) won't handle this code
    # it is not worthy to hack it for this particular piece of the code
    if sys.version_info[0] == 3 and sys.version_info[1] == 2:
        logging.error("Python 3.2 is not supported, please use other version")
        return 1

    # there are some cases (i.e. Travis CI) when this script is run on systems without DISPLAY variable being set
    # in such case matplotlib backend has to be explicitly specified
    # we do it here and not in the top of the file, as interleaving imports with code lines is discouraged
    import matplotlib
    matplotlib.use('PDF')
    import matplotlib.pyplot as plt
    from matplotlib import colors
    from matplotlib.backends.backend_pdf import PdfPages

    # parse arguments
    parser = argparse.ArgumentParser()
    parser.add_argument("spc_file", help="location of input SPC file", type=argparse.FileType('r'))
    parser.add_argument("pdf_file", help="location of output PDF file", type=argparse.FileType('w'))
    parser.add_argument('-d', '--depth', help="Plot spectra at depths",
                        type=float, nargs='+')
    parser.add_argument('-l', '--logscale', help="Enable plotting particle number on logarithmic scale",
                        action='store_true')
    parser.add_argument('-s', '--style', help="Line style", choices=['points', 'lines'], default='lines')
    parser.add_argument('-c', '--colormap', help='image color map, see http://matplotlib.org/users/colormaps.html '
                                                 'for list of possible options (default: gnuplot2)',
                        default='gnuplot2', type=str)
    parser.add_argument("-v", "--verbosity", action='count', help="increase output verbosity", default=0)
    parser.add_argument('-V', '--version', action='version', version=pt.__version__)
    parsed_args = parser.parse_args(args)

    if parsed_args.verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    elif parsed_args.verbosity > 1:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig()

    # read SPC file
    spc_object = spc.SPC(parsed_args.spc_file.name)
    spc_object.read_spc()

    # get total number of entries in SPC object, needed to allocate memory for numpy array
    n = 0
    for db in sorted(spc_object.data, key=lambda x: x.depth):
        for sb in db.species:
            n += int(sb.ne)
    logging.info("Total number of entries in SPC file {}".format(n))

    # allocate empty numpy array to store SPC data
    data = np.recarray(
        shape=(int(n), ),
        dtype=[('depth', np.float64),
               ('z', np.int8),
               ('a', np.int8),
               ('energy_left_edge', np.float64),
               ('energy_bin_width', np.float64),
               ('tot_part_number', np.float64)])
    logging.debug("Temporary numpy array to store SPC data has shape {}".format(data.shape))

    # fill numpy array with spectral data, assuming that no references are used to store bin centers
    n = 0
    for db in sorted(spc_object.data, key=lambda x: x.depth):
        for sb in sorted(db.species, key=lambda y: y.z):
            data.depth[n:n + int(sb.ne)] = db.depth
            data.z[n:n + int(sb.ne)] = sb.z
            data.a[n:n + int(sb.ne)] = sb.a
            data.energy_left_edge[n:n + int(sb.ne)] = sb.ebindata[:-1]  # left edges of bins
            data.energy_bin_width[n:n + int(sb.ne)] = np.diff(sb.ebindata)  # bin widths
            data.tot_part_number[n:n + int(sb.ne)] = sb.histdata
            n += int(sb.ne)

    logging.debug("Temporary numpy array filled with data, size {}".format(data.size))

    # find particle species, depth steps and energy bins
    az_uniq = np.unique(data[['a', 'z']])
    logging.info("Particle species A, Z: {}".format(az_uniq))

    depth_uniq = np.unique(data.depth)
    if np.unique(np.diff(depth_uniq)).size == 1:
        logging.info("{} of depth steps from {} to {}".format(len(depth_uniq), depth_uniq.min(), depth_uniq.max()))
    else:
        logging.info("Depth steps {}".format(depth_uniq))

    energy_uniq = np.unique(data.energy_left_edge)
    if np.unique(np.diff(energy_uniq)).size == 1:
        logging.info("{} of depth energy bins from {} to {} (left edges)".format(
            len(energy_uniq), energy_uniq.min(), energy_uniq.max()))
    else:
        logging.info("Energy steps {}".format(energy_uniq))

    # save multiple page PDF with plots
    with PdfPages(parsed_args.pdf_file.name) as pdf:
        logging.debug("Saving PDF file {}".format(parsed_args.pdf_file.name))
        plt.rc('text', usetex=False)

        # choose linestyle
        linestyle = '-'
        marker = ''
        if parsed_args.style == 'lines':
            linestyle = '-'
            marker = ''
        elif parsed_args.style == 'points':
            linestyle = ''
            marker = '.'

        # first a summary plot of particle number vs depth, several series corresponding to Z species
        fig, ax = plt.subplots()
        ax.set_title("Peak position {:3.3f} [cm], beam energy {} MeV/amu".format(spc_object.peakpos,
                                                                                 spc_object.energy))
        ax.set_xlabel("Depth [cm]")
        ax.set_ylabel("Particle number / primary [(none)]")
        if parsed_args.logscale:
            ax.set_yscale('log')
        for az in az_uniq:
            a, z = az
            az_data = data[(data.z == z) & (data.a == a)]  # this may cover depth only partially
            if az_data.tot_part_number.any():
                depth_steps = az_data.depth[az_data.energy_left_edge == az_data.energy_left_edge[0]]
                energy_bin_widths = az_data.energy_bin_width[az_data.depth == az_data.depth[0]]
                zlist = az_data.tot_part_number.reshape(depth_steps.size, energy_bin_widths.size).T
                part_number_at_depth = (zlist.T * energy_bin_widths).sum(axis=1)
                ax.plot(depth_steps, part_number_at_depth,
                        marker=marker, linestyle=linestyle, label="A = {}, Z = {}".format(a, z))
        plt.legend(loc=0)
        pdf.savefig(fig)
        plt.close()
        logging.debug("Summary particle number vs depth plot saved")

        if parsed_args.depth:
            # then a spectrum plot at depths selected by user
            depths_to_plot = parsed_args.depth
            if any(math.isnan(d) for d in parsed_args.depth):
                depths_to_plot = depth_uniq
                logging.info("Plotting spectrum at all {} depths requested".format(len(depths_to_plot)))
            for depth_requested in depths_to_plot:
                depth_step = data.depth[np.abs(data.depth - depth_requested).argmin()]
                logging.info("Depth step {}, depth requested {}".format(depth_step, depth_requested))
                fig, ax = plt.subplots()
                ax.set_title("Spectrum @ {:3.3f} cm".format(depth_step))
                ax.set_xlabel("Energy [MeV/amu]")
                ax.set_ylabel("Particle number / primary [1/MeV/amu]")
                if parsed_args.logscale:
                    ax.set_yscale('log')
                for az in az_uniq:
                    a, z = az
                    az_data = data[(data.z == z) & (data.a == a)]
                    if az_data.tot_part_number.any():
                        mask1 = (az_data.depth == depth_step)
                        energy_bin_centers = az_data.energy_left_edge[mask1] + 0.5 * az_data.energy_bin_width[mask1]
                        tot_part_number = az_data.tot_part_number[mask1]
                        if np.any(mask1):
                            ax.plot(energy_bin_centers, tot_part_number,
                                    linestyle=linestyle,
                                    marker=marker,
                                    label="Z = {:d} A = {:d}".format(z, a))
                plt.legend(loc=0)
                pdf.savefig(fig)
                plt.close()
                logging.debug("Spectrum for depth {} saved".format(depth_step))

        # then couple of pages, each with heatmap plot of spectrum for given particle specie
        for az in az_uniq:
            a, z = az
            az_data = data[(data.z == z) & (data.a == a)]
            if az_data.tot_part_number.any():
                depth_steps = az_data.depth[az_data.energy_left_edge == az_data.energy_left_edge[0]]
                energy_left_edges = az_data.energy_left_edge[az_data.depth == az_data.depth[0]]
                energy_bin_widths = az_data.energy_bin_width[az_data.depth == az_data.depth[0]]

                zlist = az_data.tot_part_number.reshape(depth_steps.size, energy_left_edges.size).T
                if parsed_args.logscale:
                    norm = colors.LogNorm(vmin=zlist[zlist > 0.0].min(), vmax=zlist.max())
                else:
                    norm = colors.Normalize(vmin=0.0, vmax=zlist.max())
                fig, ax = plt.subplots()
                ax.set_title("Spectrum for Z = {} A = {}".format(z, a))
                ax.set_xlabel("Depth [cm]")
                ax.set_ylabel("Energy [MeV/amu]")
                max_energy_nonzero_part_no = energy_left_edges[zlist.mean(axis=1) > 0].max()
                ax.set_ylim(0, 1.2 * max_energy_nonzero_part_no)

                # in case depth or energy steps form arithmetic progress, then pcolorfast expects
                #   third argument (Z) to be of shape len(X) times len(Y)
                # otherwise it expects
                #   third argument (Z) to be of shape len(X)-1 times len(Y)-1
                # here we check if latter is the case and add one element to X and Y sequences
                energy_steps = energy_left_edges
                depth_steps_widths = np.diff(depth_steps)
                if np.unique(depth_steps_widths).size > 1 or np.unique(energy_bin_widths).size > 1:
                    depth_steps = np.append(depth_steps, depth_steps[-1] + depth_steps_widths.min())
                    energy_steps = np.append(energy_left_edges, energy_left_edges[-1] + energy_bin_widths[-1])

                im = ax.pcolorfast(depth_steps, energy_steps, zlist, norm=norm, cmap=parsed_args.colormap)
                cbar = plt.colorbar(im)
                cbar.set_label("Particle number / primary [1/MeV/amu]", rotation=270, verticalalignment='bottom')
                pdf.savefig(fig)
                plt.close()
                logging.debug("Particle number / primary map for Z = {} saved".format(z))
            else:
                logging.warning("Skipped generation of particle number map for Z = {}, no data !".format(z))

        # File metadata
        d = pdf.infodict()
        d['Title'] = 'SPC summary for {}'.format(parsed_args.spc_file.name)
        d['Author'] = 'User {}'.format(getpass.getuser())
        d['Subject'] = '{}'.format(" ".join(sys.argv))
        d['Keywords'] = 'pytrip, spc'
        d['Creator'] = 'pytrip98 {}'.format(pt.__version__)
        d['CreationDate'] = datetime.datetime.today()
        d['ModDate'] = datetime.datetime.today()

    logging.info("Done")
    return 0


if __name__ == '__main__':
    sys.exit(main(sys.argv[1:]))



#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Module with auxiliary functions (mostly internal use).
"""

import os
import logging

logger = logging.getLogger(__name__)


def get_class_name(item):
    """
    :returns: name of class of 'item' object.
    """
    return item.__class__.__name__


class TRiP98FilePath(object):
    """
    Helper class which deals with filename discovery.
    It helps in getting stem part of the filename (i.e. patient name) from the filename which includes file extension.
    It can also construct full filename when only stem part is provided.
    Cube type can be specified in the class constructor by providing object of cube class (or the class itself).

    TRiP98 files (cubes) are named according to one of the following pattern:

    A) STEM pattern

    STEM + SUFFIX + . + hed  (header file)
    STEM + SUFFIX + . + EXTENSION  (data file)

    STEM is a core part of the path provided by user (i.e. patient name), it can end with DOT (.)
    SUFFIX is used to distinguish between quantities saved in a cube (i.e. PHYS, BIO, RBE, SVV, MLET, DOSEMLET)

    B) NAME pattern

    Here suffix is omitted and we have no information about quantity saved in a cube.
    (i.e. we can save LET cube into the files (1.hed + 1.dos); then upon reading we will not know from the filename
    whether it is a LET or DOS cube).

    NAME + . + hed (header file)
    NAME + . + EXTENSION (data file)

    Each of the files can also be gziped packed and end with .gz

    See http://bio.gsi.de/DOCS/TRiP98/PRO/DOCS/trip98cmddose.html for more details.

    """

    def __init__(self, name, cube_type):
        """
        Creates a helper class to deal with TRiP98 filenames.
        It digest full or partial filename. Header and datafile extensions are extracted from cube_type
        parameter, as each of TRiP98 cubes has a field which holds allowed extensions.
        Each cube can store different parameter (i.e. DosCube can store RBE and dose (PHYS) data.)
        List of allowed parameters is also encoded in the TRiP98 cube classes.
        :param name: full filename to header or datafile (i.e. pat19998.ctx) or part of it (i.e. pat19998).
        :param cube_type: object which identifies TRiP98 cube, i.e. object of CtxCube, DosCube or LETCube.
        """
        self.name = name
        self.cube_type = cube_type
        self.data_file_extension = self.cube_type.data_file_extension
        self.header_file_extension = self.cube_type.header_file_extension
        self.allowed_suffix = self.cube_type.allowed_suffix

    def is_valid_header_path(self):
        """
        Checks if filename corresponds to valid header filename
        (i.e. it it ends with .hed or .HED).
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.CtxCube).is_valid_header_path()
        False
        >>> TRiP98FilePath("file.txt.hed", pt.DosCube).is_valid_header_path()
        True
        >>> TRiP98FilePath("file.HED", pt.CtxCube).is_valid_header_path()
        True
        >>> TRiP98FilePath("file.DOS", pt.DosCube).is_valid_header_path()
        False
        >>> TRiP98FilePath("file.hed", pt.Cube).is_valid_header_path()
        True
        >>> TRiP98FilePath("file.hed", pt.LETCube).is_valid_header_path()
        True
        >>> TRiP98FilePath("file", pt.LETCube).is_valid_header_path()
        False

        :return: True if filename denotes header, False otherwise.
        """

        # if by some wild chance subclass of pt.Cube doesn't specify header file extension, we return False
        if not self.header_file_extension:
            return False

        # check lower and uppercase extensions
        return self._ungzipped_filename.endswith(
            (self.header_file_extension.lower(),
             self.header_file_extension.upper()))

    def is_valid_datafile_path(self):
        """
        Checks if filename corresponds to valid datafile.
        If wanted to get information about DosCube or LETCube, it checks if file extension is .dos,
        for CtxCubes it checks for .ctx extension.

        It doesn't check if filename contains a valid suffix (dosemlet or mlet for LETCube, PHYS, etc... for DosCube.
        Such type of check can be done using is_valid_cube_type() method.

        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.DosCube).is_valid_datafile_path()
        False
        >>> TRiP98FilePath("file.txt.hed", pt.DosCube).is_valid_datafile_path()
        False
        >>> TRiP98FilePath("file.dos", pt.DosCube).is_valid_datafile_path()
        True
        >>> TRiP98FilePath("file.DOS", pt.DosCube).is_valid_datafile_path()
        True
        >>> TRiP98FilePath("file.dos", pt.Cube).is_valid_datafile_path()
        False
        >>> TRiP98FilePath("file.dos", pt.LETCube).is_valid_datafile_path()
        True
        >>> TRiP98FilePath("file.dosemlet.dos", pt.DosCube).is_valid_datafile_path()
        True
        >>> TRiP98FilePath("file.dosemlet.dos", pt.LETCube).is_valid_datafile_path()
        True

        :return: True if filename denotes datafile, False otherwise.
        """

        # check if corresponding cube class has information about data file extension
        if not self.data_file_extension:
            return False

        # check lower and uppercase extensions
        correct_extension = self._ungzipped_filename.endswith((self.data_file_extension.lower(),
                                                               self.data_file_extension.upper()))

        # check if cube type can be identified using suffix part
        compatible_suffix = self.is_valid_cube_type()

        return correct_extension and compatible_suffix

    def is_valid_cube_type(self):
        no_suffix = self.suffix is None
        return no_suffix or (self.suffix in self.allowed_suffix)

    def _is_gzipped(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube)._is_gzipped()
        False
        >>> TRiP98FilePath("file.txt.gz", pt.Cube)._is_gzipped()
        True
        >>> TRiP98FilePath("file.txt.GZ", pt.Cube)._is_gzipped()
        True
        >>> TRiP98FilePath("file.txt.gZ", pt.Cube)._is_gzipped()
        False

        :return:
        """
        return self.name.endswith(('.gz', '.GZ'))

    @property
    def _ungzipped_filename(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube)._ungzipped_filename
        'file.txt'
        >>> TRiP98FilePath("file.txt.gz", pt.Cube)._ungzipped_filename
        'file.txt'
        >>> TRiP98FilePath("file.txt.GZ", pt.Cube)._ungzipped_filename
        'file.txt'
        >>> TRiP98FilePath("file.txt.gZ", pt.Cube)._ungzipped_filename
        'file.txt.gZ'

        :return:
        """
        if not self._is_gzipped():
            return self.name
        else:
            return self.name[:-3]

    @property
    def _filename_without_extension(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube)._filename_without_extension
        'file.txt'
        >>> TRiP98FilePath("file.txt.gz", pt.Cube)._filename_without_extension
        'file.txt'
        >>> TRiP98FilePath("file.txt.dos", pt.DosCube)._filename_without_extension
        'file.txt'
        >>> TRiP98FilePath("file.hed", pt.DosCube)._filename_without_extension
        'file'
        >>> TRiP98FilePath("file.txt.dos", pt.CtxCube)._filename_without_extension
        'file.txt.dos'
        """

        # remove .gz if present
        _ungzipped_filename = self._ungzipped_filename

        # check lower and uppercase extensions
        # check if header extension present, if yes, drop it
        if self.header_file_extension and _ungzipped_filename.endswith(
            (self.header_file_extension.lower(),
             self.header_file_extension.upper())):
            return self._ungzipped_filename[:-len(self.header_file_extension)]

        # check if datafile extension present, if yes, drop it
        elif self.data_file_extension and _ungzipped_filename.endswith(
            (self.data_file_extension.lower(),
             self.data_file_extension.upper())):
            return _ungzipped_filename[:-len(self.data_file_extension)]

        # no extensions found, return only unzipped filename
        else:
            return _ungzipped_filename

    @property
    def suffix(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube).suffix
        >>> TRiP98FilePath("file.txt", pt.DosCube).suffix
        >>> TRiP98FilePath("filephys.txt", pt.DosCube).suffix
        >>> TRiP98FilePath("filephys.dos", pt.DosCube).suffix
        'phys'
        >>> TRiP98FilePath("filephys.hed", pt.DosCube).suffix
        'phys'
        >>> TRiP98FilePath("filephys.hed", pt.LETCube).suffix
        >>> TRiP98FilePath("filemlet.hed", pt.LETCube).suffix
        'mlet'
        >>> TRiP98FilePath("file.mlet.hed", pt.LETCube).suffix
        'mlet'

        :return:
        """
        _filename_without_extension = self._filename_without_extension
        if not self.allowed_suffix:
            return None
        for suffix in self.allowed_suffix:
            if _filename_without_extension.endswith((suffix.lower(), suffix.upper())):
                return _filename_without_extension[-len(suffix):]
        return None

    @property
    def stem(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube).stem
        >>> TRiP98FilePath("file.txt", pt.DosCube).stem
        >>> TRiP98FilePath("filephys.txt", pt.DosCube).stem
        >>> TRiP98FilePath("filephys.dos", pt.DosCube).stem
        'file'
        >>> TRiP98FilePath("filephys.hed", pt.DosCube).stem
        'file'
        >>> TRiP98FilePath("filephys.hed", pt.LETCube).stem
        >>> TRiP98FilePath("filemlet.hed", pt.LETCube).stem
        'file'
        >>> TRiP98FilePath("file.mlet.hed", pt.LETCube).stem
        'file.'

        :return:
        """
        if self._is_stem_pattern():
            return self._filename_without_extension[:-len(self.suffix)]
        else:
            return None

    def _is_stem_pattern(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("file.txt", pt.Cube)._is_stem_pattern()
        False
        >>> TRiP98FilePath("file.txt", pt.DosCube)._is_stem_pattern()
        False
        >>> TRiP98FilePath("filephys.txt", pt.DosCube)._is_stem_pattern()
        False
        >>> TRiP98FilePath("filephys.dos", pt.DosCube)._is_stem_pattern()
        True
        >>> TRiP98FilePath("filephys.hed", pt.DosCube)._is_stem_pattern()
        True
        >>> TRiP98FilePath("filephys.hed", pt.LETCube)._is_stem_pattern()
        False
        >>> TRiP98FilePath("filemlet.hed", pt.LETCube)._is_stem_pattern()
        True
        >>> TRiP98FilePath("file.mlet.hed", pt.LETCube)._is_stem_pattern()
        True

        :return:
        """
        return bool(self.suffix)

    @property
    def datafile(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("patient1", pt.DosCube).datafile
        'patient1.dos'
        >>> TRiP98FilePath("patient1.dos", pt.DosCube).datafile
        'patient1.dos'
        >>> TRiP98FilePath("patient1.hed", pt.DosCube).datafile
        'patient1.dos'
        >>> TRiP98FilePath("patient1.phys.hed", pt.DosCube).datafile
        'patient1.phys.dos'

        :return:  path to the (unzipped) data file.
        """
        if self.is_valid_header_path():
            return self._filename_without_extension + self.data_file_extension
        elif self.is_valid_datafile_path():
            return self.name
        else:
            return self.name + self.data_file_extension

    @property
    def header(self):
        """
        >>> import pytrip as pt
        >>> TRiP98FilePath("patient1", pt.DosCube).header
        'patient1.hed'
        >>> TRiP98FilePath("patient1.dos", pt.DosCube).header
        'patient1.hed'
        >>> TRiP98FilePath("patient1.hed", pt.DosCube).header
        'patient1.hed'
        >>> TRiP98FilePath("patient1.phys.hed", pt.DosCube).header
        'patient1.phys.hed'

        :return:  path to the (unzipped) data file.
        """
        if self.is_valid_header_path():
            return self.name
        elif self.is_valid_datafile_path():
            return self._filename_without_extension + self.header_file_extension
        else:
            return self.name + self.header_file_extension

    @property
    def basename(self):
        """
        Returns the basename, without leading directory or trailing suffixes.

        >>> import pytrip as pt
        >>> TRiP98FilePath("foobar/patient1", pt.DosCube).basename
        'patient1'
        >>> TRiP98FilePath("patient1.dos", pt.DosCube).basename
        'patient1'
        >>> TRiP98FilePath("patient1.dos", pt.CtxCube).basename
        'patient1.dos'
        >>> TRiP98FilePath("patient1.dos.gz", pt.DosCube).basename
        'patient1'
        >>> TRiP98FilePath("patient1.hed", pt.DosCube).basename
        'patient1'
        >>> TRiP98FilePath("patient1.phys.hed", pt.DosCube).basename
        'patient1.phys'

        :return:  path to the (unzipped) data file.
        """
        # return self._filename_without_extension
        return os.path.basename(self._filename_without_extension)

    @property
    def dir_basename(self):
        """
        Returns the basename including the leading directory.

        >>> import pytrip as pt
        >>> TRiP98FilePath("foobar/patient1", pt.DosCube).dir_basename
        'foobar/patient1'
        >>> TRiP98FilePath("patient1.dos", pt.DosCube).dir_basename
        'patient1'
        >>> TRiP98FilePath("patient1.dos", pt.CtxCube).dir_basename
        'patient1.dos'
        >>> TRiP98FilePath("patient1.dos.gz", pt.DosCube).dir_basename
        'patient1'
        >>> TRiP98FilePath("patient1.hed", pt.DosCube).dir_basename
        'patient1'
        >>> TRiP98FilePath("patient1.phys.hed", pt.DosCube).dir_basename
        'patient1.phys'

        :return:  path + basename to the (unzipped) data file.
        """
        # return self._filename_without_extension
        return self._filename_without_extension


class TRiP98FileLocator(object):
    def __init__(self, name, cube_type):
        self.trip98path = TRiP98FilePath(name, cube_type)

        # construct list of suffixes to check
        # they may or may not include leading dot (i.e. ".dosemlet" or "dosemlet")
        # suffix may also be lower or UPPERCASE
        self.list_of_suffixes_to_check = []
        for optional_dot in ('', '.'):
            for suffix in self.trip98path.allowed_suffix:
                self.list_of_suffixes_to_check.append(optional_dot + suffix.lower())
                self.list_of_suffixes_to_check.append(optional_dot + suffix.upper())
        self.list_of_suffixes_to_check.append("")

    @property
    def header(self):
        """
        Calculates the path to the header file existing on the disk.
        Assuming that in the filesystem files "a.hed" and "c.ctx" are present it will work like following:

        >>> import pytrip as pt
        >>> TRiP98FileLocator("tests/res/TST003/tst003000", pt.CtxCube).header
        'tests/res/TST003/tst003000.hed'

        It can also work well with gzipped filenames:
        >>> ungzipped_file = "tests/res/TST003/tst003012.dosemlet.hed"
        >>> os.path.exists(ungzipped_file)
        False

        >>> gzipped_file = ungzipped_file + ".gz"
        >>> gzipped_file
        'tests/res/TST003/tst003012.dosemlet.hed.gz'
        >>> os.path.exists(gzipped_file)
        True

        FileLocator can automatically find requested file by adding suffix dosemlet which is appropriate to
        requested cube (note that the returned path points to the file which exists on the disk).
        >>> TRiP98FileLocator('tests/res/TST003/tst003012', pt.LETCube).header
        'tests/res/TST003/tst003012.dosemlet.hed.gz'

        After changing the cube to DosCube we get the path to another file (also located on the disk).
        >>> TRiP98FileLocator('tests/res/TST003/tst003012', pt.DosCube).header
        'tests/res/TST003/tst003012.hed'

        :return: path to the header file which exists on the filesystem or None if not found
        """
        dir_basename = self.trip98path.dir_basename
        files_tried = []
        logger.info("Locating : " + self.trip98path.name + " as " + str(self.trip98path.cube_type))

        for suffix in self.list_of_suffixes_to_check:
            for gzip_extension in ("", ".gz", ".GZ"):
                for header_extension in (self.trip98path.header_file_extension.lower(),
                                         self.trip98path.header_file_extension.upper()):
                    candidate_path = dir_basename + suffix + header_extension + gzip_extension
                    if os.path.exists(candidate_path):
                        logger.info("Found " + candidate_path)
                        return candidate_path
                    else:
                        files_tried.append(candidate_path)
        logger.warning("Checking following files: " + " , ".join(files_tried) + ". None of them exists.")
        return None

    @property
    def datafile(self):
        """
        It works exactly in the same way as header method in TRiP98FileLocator class,
        but returns path to the datafile.

        :return: path to the data file which exists on the filesystem or None if not found
        """
        dir_basename = self.trip98path.dir_basename
        files_tried = []
        logger.info("Locating : " + self.trip98path.name + " as " + str(self.trip98path.cube_type))
        for suffix in self.list_of_suffixes_to_check:
            for gzip_extension in ("", ".gz", ".GZ"):
                for datafile_extension in (self.trip98path.data_file_extension.lower(),
                                           self.trip98path.data_file_extension.upper()):
                    candidate_path = dir_basename + suffix + datafile_extension + gzip_extension
                    if os.path.exists(candidate_path):
                        return candidate_path
                    else:
                        files_tried.append(candidate_path)
        logger.warning("Checking following files: " + " , ".join(files_tried) + ". None of them exists.")
        return None


def evaluator(funct, name='funct'):
    """ Wrapper for evaluating a function.

    :params str funct: string which will be parsed
    :params str name: name which will be assigned to created function.

    :returns: function f build from 'funct' input.
    """
    code = compile(funct, name, 'eval')

    def f(x):
        return eval(code, locals())

    f.__name__ = name
    return f



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
TODO: documentation here.
"""
import os
import setuptools
import subprocess

import numpy as np


def git_version():
    """
    Inspired by https://github.com/numpy/numpy/blob/master/setup.py
    :return: the git revision as a string
    """
    def _minimal_ext_cmd(cmd):
        # construct minimal environment
        env = {}
        for k in ['SYSTEMROOT', 'PATH', 'HOME']:
            v = os.environ.get(k)
            if v is not None:
                env[k] = v
        # LANGUAGE is used on win32
        env['LANGUAGE'] = 'C'
        env['LANG'] = 'C'
        env['LC_ALL'] = 'C'
        out = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env).communicate()[0]
        return out

    try:
        out = _minimal_ext_cmd(['git', 'describe', '--tags', '--long'])
        GIT_REVISION = out.strip().decode('ascii')
        print('GIT_REVISION', GIT_REVISION)
        if GIT_REVISION:
            no_of_commits_since_last_tag = int(GIT_REVISION.split('-')[1])
            tag_name = GIT_REVISION.split('-')[0][1:]
            if no_of_commits_since_last_tag == 0:
                version = tag_name
            else:
                version = '{}+rev{}'.format(tag_name, no_of_commits_since_last_tag)
        else:
            version = "Unknown"
    except OSError:
        version = "Unknown"

    return version


def write_version_py(filename='pytrip/__init__.py'):
    cnt = """
__version__ = '%(version)s'
"""

    GIT_REVISION = git_version()
    a = open(filename, 'a')
    try:
        a.write(cnt % {'version': GIT_REVISION})
    finally:
        a.close()


write_version_py()

with open('README.rst') as readme_file:
    readme = readme_file.read()

extensions = [setuptools.Extension(
    'pytriplib',
    sources=[os.path.join('pytrip', 'lib', 'core.c')],
    extra_compile_args=['-fpic']),
    setuptools.Extension(
        '_cntr',
        sources=[os.path.join('pytrip', 'lib', 'cntr.c')],
        extra_compile_args=['-fpic'])
]

setuptools.setup(
    name='pytrip98',
    version=git_version(),
    packages=setuptools.find_packages(exclude="tests"),
    url='https://github.com/pytrip/pytrip',
    license='GPL',
    author='Jakob Toftegaard, Niels Bassler, Leszek Grzanka',
    author_email='leszek.grzanka@gmail.com',
    description='PyTRiP',
    long_description=readme + '\n',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Medical Science Apps.',
        'Topic :: Scientific/Engineering :: Physics',

        # OS and env
        'Environment :: Console',
        'Operating System :: POSIX :: Linux',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: C',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython'
    ],
    package_data={'pytrip': ['data/*.dat', 'pytriplib.*', 'cntr.*']},
    install_requires=[
        'matplotlib', 'numpy', 'pydicom'
    ],
    include_dirs=[np.get_include()],
    ext_modules=extensions,
    entry_points={
        'console_scripts': [
            'trip2dicom=pytrip.utils.trip2dicom:main',
            'dicom2trip=pytrip.utils.dicom2trip:main',
            'cubeslice=pytrip.utils.cubeslice:main',
            'rst2sobp=pytrip.utils.rst2sobp:main',
            'rst_plot=pytrip.utils.rst_plot:main',
            'bevlet2oer=pytrip.utils.bevlet2oer:main',
            'gd2dat=pytrip.utils.gd2dat:main',
            'gd2agr=pytrip.utils.gd2agr:main',
            'spc2pdf=pytrip.utils.spc2pdf:main',
        ],
    },
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.3.*'
)



#
#    Copyright (C) 2010-2018 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Field() objects here, which can be passed to a Plan() object.
"""

import logging
import uuid

logger = logging.getLogger(__name__)


class Field(object):
    """ One or more Field() object, which then can be added to a Plan() object.
    :params str basename: basename of field without file extension (input or output will be suffixed with
    proper file extension)
    """

    def __init__(self, basename=""):
        """ Create an instance of the Field class.
        :params str basename" The name of this field, will be used as basename for .rst files.
        """
        self.__uuid__ = uuid.uuid4()  # for uniquely identifying this field
        # basename of the field (i.e. without file extension).
        # Any input/output process will be suffixed with .rst
        self.basename = basename
        self.number = 1  # Field number. First field must be 1
        self.gantry = 0.0  # TRiP98 angles assumed here.
        self.couch = 0.0  # TRiP98 angles assumed here.
        self.fwhm = 4.0  # spot width in [mm]
        self.raster_step = [2, 2]  # spot distance in [mm]
        self.dose_extension = 1.2  # see TRiP98 field / doseext() documentation
        self.contour_extension = 0.6  # see TRiP98 field / contourext() documentation

        # the field / command in TRiP98 may take an external rst file as an input, and will then
        # instead of optimization do a forward calculation based on this file.
        # The filename will be constructed from the basename given for this field.
        # However, we need to tell, that we provide the rst.
        self.use_raster_file = False

        self.zsteps = 1.0  # in [mm]
        # the next two are deprecated, use KernelModel
        self.kernel = None

        # isocenter holds the [x,y,z] coordinates of the isocenter/target in [mm].
        # This is used for the field / target() option of the TRiP98
        # It can be used to override the automatically calculated isocenter from TRiP98.
        self.isocenter = []

        self.selected = False

    def __str__(self):
        """ str output handler
        """
        return self._print()

    def _print(self):
        """ Pretty print all attributes.
        """
        out = "\n"
        out += "   Field {:d} '{:s}'\n".format(self.number, self.basename)
        out += "----------------------------------------------------------------------------\n"
        out += "|  UUID                         : {:s}\n".format(str(self.__uuid__))
        out += "|  Couch angle                  : {:.2f} deg\n".format(self.couch)
        out += "|  Gantry angle                 : {:.2f} deg\n".format(self.gantry)
        out += "|\n"
        out += "|  Spot size (FWHM)             : {:.2f} mm\n".format(self.fwhm)
        out += "|  Raster step size (x,y)       : {:.2f}, {:.2f} mm\n".format(self.raster_step[0],
                                                                              self.raster_step[1])
        out += "|  Z-steps                      : {:.2f} mm\n".format(self.zsteps)
        out += "|\n"
        out += "|  Dose extension               : {:.2f}\n".format(self.dose_extension)
        out += "|  Contour extension            : {:.2f}\n".format(self.contour_extension)
        out += "|  Use external .rst file       : {:s}\n".format(str(self.use_raster_file))
        if self.isocenter:
            out += "|  Isocenter (x,y,z)            : {:.2f} {:.2f} {:.2f} mm\n".format(self.isocenter[0],
                                                                                        self.isocenter[1],
                                                                                        self.isocenter[2])
        else:
            out += "|  Isocenter (x,y,z)            : (not set)\n"
        out += "----------------------------------------------------------------------------\n"
        return out

    def set_isocenter_from_string(self, isocenter_str):
        """ Override the automatically determined isocenter from TRiP98.
        :params str isocenter_str: x,y,z coordinates of the isocenter/target in [mm] in a comma delimted string.
        such as "123.33,158.4,143.5".
        If and empty string is provided, then the isocenter is unset, and TRiP98 will calculate it itself.

        Following the target() option in the TRiP98 field command, one can specify
        the isocenter of the target. If not used, TRiP98 will determine the isocenter
        from the target Voi provided.

        """
        if len(isocenter_str) == 0:
            self.isocenter = []
            return
        _target = isocenter_str.split(",")
        if len(_target) == 3:
            try:
                self.isocenter = [float(_target[0]), float(_target[1]), float(_target[2])]
                return
            except Exception:
                logger.error("Expected a 'X,Y,Z' formatted string for Field().set_isocenter_from_string")



#
#    Copyright (C) 2010-2017 PyTRiP98 Developers.
#
#    This file is part of PyTRiP98.
#
#    PyTRiP98 is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    PyTRiP98 is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with PyTRiP98.  If not, see <http://www.gnu.org/licenses/>.
#
"""
Additional tools which are not tied to any particular model.
"""
import numpy as np
import logging

logger = logging.getLogger(__name__)


def rbe_from_sf(sf_ion, dose_ion, alpha_x, beta_x):
    """
    Returns the RBE for given ion survivng fraction and (alpha/beta)x-ray
    :params float dose: ion physical dose in [Gy] (cube or scalar)
    :params float sf_ion: surviving fraction in ion beam (cube of scalar)
    :params float alpha_x: alpha for X-rays in [Gy^-1] (cube or scalar)
    :params float beta_x: beta for X-rays in [Gy^-2] (cube or scalar)
    """

    # Solve for D_x: beta_x*D*2 + alpha_x*D + ln(sf_ion) = 0
    # RBE = D_x / D_ion

    a = beta_x * dose_ion * 2.0
    b = alpha_x * dose_ion
    c = np.log(sf_ion)  # natural logarithm
    d = b * b - 4 * a * c
    x1 = (-b + np.sqrt(d)) / (2.0 * a)
    x2 = (-b - np.sqrt(d)) / (2.0 * a)

    if x1 > x2:
        dose_x = x1
    else:
        dose_x = x2

    rbe = dose_x / dose_ion

    if rbe < 0:
        logger.warning("negative rbe encountered")
    return rbe


def lq(dose_x, alpha_x, beta_x):
    """
    Linear-quadratic survival model. (LQ-model)

    Returns surviving fraction for a given dose, alpha and beta for X-rays.

    :params float dose: x-ray physical dose in [Gy] (cube or scalar)
    :params float alpha_x: alpha value for x-rays [Gy^-1] (cube or scalar)
    :params float beta_x: beta value for x-rays [Gy^-2] (cube or scalar)
    """
    return np.exp(-alpha_x * dose_x - beta_x * dose_x * dose_x)
